import collections
import fnmatch
import re
import reprlib
import sys
from inspect import isclass
import GangaCore.Utility.logging
from GangaCore.Core.exceptions import GangaException
from GangaCore.Core.GangaRepository.Registry import RegistryKeyError, RegistryIndexError, RegistryAccessError, IncompleteObject

from GangaCore.GPIDev.Schema import ComponentItem
from GangaCore.Utility.external.OrderedDict import OrderedDict as oDict
import GangaCore.Utility.Config
from GangaCore.GPIDev.Base.Proxy import isType, stripProxy, getName, addProxy
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.Config import makeConfig

logger = getLogger()

config = GangaCore.Utility.Config.getConfig('Display')

class RegistrySlice(object):

    def __init__(self, name, display_prefix):
        """
        Constructor for a Registry Slice object
        Args:
            name (str): Slightly descriptive name of the slice
            display_prefix (str): Prefix for slice to help modify display and other string reps of the object
        """
        super(RegistrySlice, self).__init__()
        self.objects = oDict()
        self.name = name
        self._display_prefix = display_prefix
        self._display_columns = config[self._display_prefix + '_columns']
        self._display_columns_show_empty = config[self._display_prefix + "_columns_show_empty"]
        self._display_columns_width = config[self._display_prefix + "_columns_width"]
        self._display_columns_functions = {}
        try:
            col_funcs = config[self._display_prefix + '_columns_functions']
            for this_col_func in col_funcs:
                self._display_columns_functions[this_col_func] = eval(col_funcs[this_col_func])
        except Exception as x:
            logger.error("Error on evaluating display column functions from config file: %s: %s" % (getName(x), x))

        from GangaCore.Utility.ColourText import Effects
        self._colour_normal = Effects().normal
        self._proxyClass = None

    def _getColour(self, obj):
        """ Override this function in derived slices to colorize your job/task/... list"""
        return self._colour_normal

    def do_collective_operation(self, keep_going, method, *args, **kwds):
        """
        """
        if not isinstance(keep_going, bool):
            raise GangaException("The variable 'keep_going' must be a boolean. Probably you wanted to do %s(%s).%s()" % (self.name, keep_going, method))
        result = []
        for _id in self.objects.keys():
            obj = self.objects[_id]
            try:
                if isinstance(method, str):
                    doc = method
                    result.append(getattr(obj, method)(*args, **kwds))
                else:
                    try:
                        doc = method.__doc__
                    except AttributeError:
                        doc = str(method)
                    result.append(method(obj, *args, **kwds))
            except GangaException as x:
                if not keep_going:
                    raise
            except Exception as x:
                logger.exception('%s %s %s: %s %s', doc, self.name, _id, getName(x), x)
                if not keep_going:
                    raise
        return result

    def ids(self, minid=None, maxid=None):
        "Get the list of job ids. 'minid' and 'maxid' specify optional (inclusive) slice range."
        if maxid is None:
            maxid = sys.maxsize
        if minid is None:
            minid = 0
        return [k for k in self.objects.keys() if minid <= k <= maxid]

    def clean(self, confirm=False, force=False):
        """Cleans the repository only if this slice represents the repository
        Returns True on success and False on failure"""
        if not hasattr(self.objects, "clean"):
            logger.error("'clean' only works on whole registries, e.g. 'jobs.clean()'. Use remove() to delete job slices")
            return False
        if not confirm:
            logger.warning("You are about to irreversibly delete the WHOLE '%s' registry, without properly cleaning up individual jobs." % (self.objects.name))
            if force:
                logger.warning("You will also cause any other Ganga sessions accessing this repository to shut down their operations")
                logger.warning("If you just want to remove all jobs, type '%s.remove()'. If you really want to do this, type '%s.clean(confirm=True,force=True)" % (self.objects.name, self.objects.name))
            else:
                logger.warning("If you just want to remove all jobs, type '%s.remove()'. If you really want to do this, type '%s.clean(confirm=True)" % (self.objects.name, self.objects.name))
            return False
        return self.objects.clean(force)

    def select(self, minid=None, maxid=None, **attrs):
        from GangaCore.GPIDev.Lib.Job.Job import Job

        if isType(minid, Job):
            if minid.master is not None:
                minid = minid.master.id
            else:
                minid = minid.id
            if maxid is None:
                maxid = minid

        if isType(maxid, Job):
            if maxid.master is not None:
                maxid = maxid.master.id
            else:
                maxid = maxid.id

        logger = getLogger()

        this_repr = reprlib.Repr()
        from GangaCore.GPIDev.Base.Proxy import addProxy
        attrs_str = ""
        ## Loop through all possible input combinations to constructa string representation of the attrs from possible inputs
        ## Required to flatten the additional arguments into a flat string in attrs_str
        for a in attrs:
            if isclass(attrs[a]):
                this_attr = addProxy(attrs[a]())
            else:
                from GangaCore.GPIDev.Base.Objects import GangaObject
                if isType(attrs[a], GangaObject):
                    this_attr = addProxy(attrs[a])
                else:
                    if type(attrs[a]) is str:
                        from GangaCore.GPIDev.Base.Proxy import getRuntimeGPIObject
                        this_attr = getRuntimeGPIObject(attrs[a], True)
                    else:
                        this_attr = attrs[a]
            full_str = str(this_attr)
            split_str = full_str.split('\n')
            for line in split_str:
                line = line.strip()
            flat_str = ''.join(split_str)
            attrs_str += ", %s=\"%s\"" % (a, flat_str)

        logger.debug("Attrs_Str: %s" % attrs_str)
        logger.debug("Constructing slice: %s" % ("%s.select(minid='%s', maxid='%s'%s)" % (self.name, this_repr.repr(minid), this_repr.repr(maxid), attrs_str)))
        this_slice = self.__class__("%s.select(minid='%s', maxid='%s'%s)" % (self.name, this_repr.repr(minid), this_repr.repr(maxid), attrs_str))

        def append(id, obj):
            this_slice.objects[id] = obj
        self.do_select(append, minid, maxid, **attrs)
        return this_slice

    def do_select(self, callback, minid=None, maxid=None, **attrs):
        """Get the slice of jobs. 'minid' and 'maxid' specify optional (inclusive) slice range.
        The returned slice object has the job registry interface but it is not connected to
        persistent storage. 
        """

        logger = getLogger()

        ## Loop through attrs to parse possible inputs into instances of a class where appropriate
        ## Unlike the select method we need to populate this dictionary with instance objects, not str or class
        for k, v in attrs.items():
            if isclass(v):
                attrs[k] = v()
            elif type(attrs[k]) is str:
                from GangaCore.GPIDev.Base.Proxy import getRuntimeGPIObject
                new_val = getRuntimeGPIObject(attrs[k], True)
                if new_val is None:
                    continue
                if isclass(new_val):
                    attrs[k] = new_val()
                else:
                    attrs[k] = new_val

        logger.debug("do_select: attrs: %s" % attrs)

        def select_by_list(this_id):
            return this_id in ids

        def select_by_range(this_id):
            return minid <= this_id <= maxid

        ids = None

        if isinstance(minid, collections.abc.Container):
            ids = minid
            select = select_by_list
        else:
            if minid is None:
                minid = 0
            if maxid is None:
                maxid = sys.maxsize
            select = select_by_range

        for this_id in self.objects.keys():
            obj = self.objects[this_id]
            logger.debug("id, obj: %s, %s" % (this_id, obj))
            if select(int(this_id)):
                logger.debug("Selected: %s" % this_id)
                selected = True
                if self.name == 'box':
                    name_str = obj._getRegistry()._getName(obj)
                else:
                    name_str = ''
                for a in attrs:
                    if self.name == 'box':
                        attrvalue = attrs[a]
                        if a == 'name':
                            if not fnmatch.fnmatch(name_str, attrvalue):
                                selected = False
                                break
                        elif a == 'application':
                            if hasattr(obj, 'application'):
                                if not getName(obj.application) == attrvalue:
                                    selected = False
                                    break
                            else:
                                selected = False
                                break
                        elif a == 'type':
                            if not getName(obj) == attrvalue:
                                selected = False
                                break
                        else:
                            from GangaCore.GPIDev.Base import GangaAttributeError
                            raise GangaAttributeError(
                                'undefined select attribute: %s' % a)
                    else:

                        if a == 'ids':
                            if int(this_id) not in attrs['ids']:
                                selected = False
                                break
                        else:
                            try:
                                item = obj._schema.getItem(a)
                                logger.debug("Here: %s, is item: %s" % (a, type(item)))
                            except KeyError as err:
                                from GangaCore.GPIDev.Base import GangaAttributeError
                                logger.debug("KeyError getting item: '%s' from schema" % a)
                                raise GangaAttributeError('undefined select attribute: %s' % a)
                            else:
                                attrvalue = attrs[a]

                                if item.isA(ComponentItem):
                                    ## TODO we need to distinguish between passing a Class type and a defined class instance
                                    ## If we passed a class type to select it should look only for classes which are of this type
                                    ## If we pass a class instance a compartison of the internal attributes should be performed
                                    from GangaCore.GPIDev.Base.Filters import allComponentFilters

                                    cfilter = allComponentFilters[item['category']]
                                    filtered_value = cfilter(attrs[a], item)
                                    from GangaCore.GPIDev.Base.Proxy import getName
                                    if not filtered_value is None:
                                        attrvalue = getName(filtered_value)
                                    else:
                                        attrvalue = getName(attrvalue)

                                    if getName(getattr(obj, a)) != attrvalue:
                                        selected = False
                                        break
                                else:
                                    if isinstance(attrvalue, str):
                                        regex = fnmatch.translate(attrvalue)
                                        reobj = re.compile(regex)
                                        # Compare the type of the attribute
                                        # against attrvalue
                                        if not reobj.match(str(getattr(obj, a))):
                                            selected = False
                                    else:
                                        if getattr(obj, a) != attrvalue:
                                            selected = False
                                            break
                if selected:
                    logger.debug("Actually Selected")
                    callback(this_id, obj)
                else:
                    logger.debug("NOT Actually Selected")
            else:
                logger.debug("NOT Selected: %s" % this_id)

    def copy(self, keep_going):
        this_slice = self.__class__("copy of %s" % self.name)
        for _id in self.objects.keys():
            obj = self.objects[_id]
            #obj = _unwrap(obj)
            copy = obj.clone()
            # If the copied object is not automatically registered,
            # try to register it in the old objects registry
            new_id = copy._getRegistryID()
            if new_id is None:
                reg = obj._getRegistry()
                if reg is None:
                    new_id = _id
                else:
                    reg._add(copy)
                    new_id = copy._getRegistryID()
            this_slice.objects[new_id] = copy
        return this_slice

    def __contains__(self, j):
        return j.id in self.objects.keys()

    def __call__(self, this_id):
        """ Retrieve an object by id.
        """
        if isinstance(this_id, str):
            if this_id.isdigit():
                this_id = int(this_id)
            else:
                matches = [o for o in self.objects if fnmatch.fnmatch(o._getRegistry()._getName(o), this_id)]
                if len(matches) > 1:
                    logger.error('Multiple Matches: Wildcards are allowed for ease of matching, however')
                    logger.error('                  to keep a uniform response only one item may be matched.')
                    logger.error('                  If you wanted a slice, please use the select method')
                    raise RegistryKeyError("Multiple matches for id='%s':%s" % (this_id, str([x._getRegistry()._getName(x) for x in matches])))
                if len(matches) < 1:
                    return
                return addProxy(matches[0])
        try:
            return addProxy(self.objects[this_id])
        except KeyError as err:
            logger.debug('Object id=%d not found' % this_id)
            logger.debug("%s" % err)
            raise RegistryKeyError('Object id=%d not found' % this_id)

    def __iter__(self):
        "Iterator for the objects. "
        class Iterator(object):

            def __init__(self, reg):
                self.it = list(reg.objects.values()).__iter__()

            def __iter__(self): return self

            def __next__(self):
                return next(self.it)
        return Iterator(self)

    def __len__(self):
        "Number of objects in the registry"
        return len(self.objects)

    def __getitem__(self, x):
        """Retrieve the job object from the registry: registry[x].
         If 'x' is a job id (int) then a single job object is returned or IndexError.
         If 'x' is a name (string) then a unique same name is returned, otherwise [].
         If 'x' is a job object then it is returned if it belongs to the registry, otherwise None.
         If 'x' is not of any of the types above, raise TypeError.
          or by name. If retrieved by name then the job must be unique, otherwise the RegistryKeyError is raised.
         If the input is incorrect, RegistryAccessError is raised.
        """
        if isinstance(x, int):
            try:
                if x < 0:
                    return addProxy(self.objects[self.ids()[x]])
                else:
                    return addProxy(self.objects[x])
            except IndexError:
                raise RegistryIndexError('list index out of range')

        if isinstance(x, str):
            ids = []
            for i in self.ids():
                j = self.objects[i]
                if j.name == x:
                    ids.append(j.id)
            if len(ids) > 1:
                raise RegistryKeyError('object "%s" not unique' % x)
            if len(ids) == 0:
                raise RegistryKeyError('object "%s" not found' % x)
            return addProxy(self.objects[ids[0]])

        if isinstance(x, slice):
            start = x.start if x.start is not None else ''
            stop = x.stop if x.stop is not None else ''

            returnable = self.__class__("%s[%s:%s]" % (self.name, start, stop))
            for id_ in self.ids()[x]:
                returnable.objects[id_] = self.objects[id_]

            return addProxy(returnable)

        raise RegistryAccessError('Expected int or string (job name).')

    @staticmethod
    def _getatr(obj, members):
        val = getattr(obj, members[0])
        if len(members) > 1:
            return str(RegistrySlice._getatr(val, members[1:]))
        else:
            return str(val)

    def _get_display_value(self, _obj, item):
        try:
            obj = stripProxy(_obj)
            try:
                if item in self._display_columns_functions:
                    display_func = self._display_columns_functions[item]
                    val = display_func(obj)
                else:
                    val = self._getatr(obj, item.split('.'))
            except KeyError as err:
                logger.debug("_get_display_value KeyError: %s" % err)
                logger.debug("item: \"%s\"" % item)
                #logger.debug("func: %s" % config[self._display_prefix + '_columns_functions'])
                #val = self._getatr(obj, item.split('.'))
                val = ""
            if not val and not item in self._display_columns_show_empty:
                val = ""
        except AttributeError as err:
            logger.debug("AttibErr: %s" % err)
            val = ""
        finally:
            pass
        return str(val)

    def _display(self, interactive=0):
        from GangaCore.Utility.ColourText import ANSIMarkup, NoMarkup, Effects
        if interactive:
            markup = ANSIMarkup()
        else:
            markup = NoMarkup()

        # default column width
        default_width = 10

        cnt = len(self)
        ds = "Registry Slice: %s (%d objects)\n" % (self.name, cnt)

        this_format = "#"
        flist = []
        for d in self._display_columns:
            width = self._display_columns_width.get(d, default_width)
            flist.append("%" + str(width) + "s ")
            #this_format += "%"+str(width)+"s  "
        this_format = "|".join(flist)
        this_format += "\n"

        if cnt > 0:
            ds += "--------------\n"
            ds += this_format % self._display_columns
            ds += "-" * len(this_format % tuple([""] * len(self._display_columns))) + "\n"


        if hasattr(self.objects, '_private_display'):
            ds += self.objects._private_display(self, this_format, default_width, markup)

        else:
            for obj_i in self.ids():
                if isinstance(self.objects[obj_i], IncompleteObject):
                    continue

                cached_data = None

                reg_object = self.objects[obj_i]
                obj = stripProxy(reg_object)
                colour = self._getColour(obj)

                vals = []
                for item in self._display_columns:
                    display_str = "display:" + str(item)
                    #logger.info("Looking for : %s" % display_str)
                    width = self._display_columns_width.get(item, default_width)
                    try:
                        if item == "fqid":
                            vals.append(str(obj._index_cache[display_str]))
                        else:
                            vals.append(str(obj._index_cache[display_str])[0:width])
                        continue
                    except KeyError as err:
                        logger.debug("_display KeyError: %s" % err)
                        #pass
                        if item == "fqid":
                            vals.append(self._get_display_value(obj, item))
                        elif item == 'subjob status':
                            vals.append('---') #As this is new we don't want to reload everything
                        else:
                            vals.append(self._get_display_value(obj, item)[0:width])
                ds += markup(this_format % tuple(vals), colour)

        return ds

    __str__ = _display

    def _id(self):
        return id(self)




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: File.py,v 1.2 2008-09-09 14:37:16 moscicki Exp $
##########################################################################

from GangaCore.Core.exceptions import GangaException, SchemaError
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, GangaFileItem
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.GPIDev.Base.Proxy import stripProxy, GPIProxyObjectFactory
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.Core.GangaRepository.VStreamer import to_file, from_file
from GangaCore.Core.GangaRepository.GangaRepositoryXML import safe_save
from GangaCore.GPIDev.Base.Objects import synchronised
import os
import shutil
import uuid

from GangaCore.Utility.files import expandfilename, chmod_executable, is_executable

from GangaCore.Utility.logging import getLogger

from GangaCore.GPIDev.Base.Filters import allComponentFilters

import re

import GangaCore.Utility.Config

from GangaCore.GPIDev.Lib.File import getSharedPath

from GangaCore.Runtime.GPIexport import exportToGPI

import threading

# regex [[PROTOCOL:][SETYPE:]..[<alfanumeric>:][/]]/filename
urlprefix = re.compile(r'^(([a-zA-Z_][\w]*:)+/?)?/')

logger = getLogger()

_prepare_lock = threading.RLock()

class File(GangaObject):

    """Represent the files, both local and remote and provide an interface to transparently get access to them.

    Typically in the context of job submission, the files are copied to the directory where the application
    runs on the worker node. The 'subdir' attribute influances the destination directory. The 'subdir' feature
    is not universally supported however and needs a review.

    """
    _schema = Schema(Version(1, 1), {'name': SimpleItem(defvalue="", doc='path to the file source'),
                                     'subdir': SimpleItem(defvalue=os.curdir, doc='destination subdirectory (a relative path)'),
                                     'executable': SimpleItem(defvalue=False, hidden=True, transient=True, doc='specify if executable bit should be set when the file is created (internal framework use)')})
    _category = 'files'
    _name = "File"
    _exportmethods = ["getPathInSandbox", "exists", "create", "isExecutable"]

    # added a subdirectory to the File object. The default is os.curdir, that is "." in Unix.
    # The subdir is a relative path and will be appended to the pathname when writing out files.
    # Therefore changing subdir to a anything starting with "/" will still end up relative
    # to the pathname when the file is copied.
    #
    # There is no protection on putting the parent directory. So ".." is legal and will make
    # the file end up in the parent directory. - AM
    def __init__(self, name=None, subdir=os.curdir):
        super(File, self).__init__()

        if not name is None:
            try:
                assert(isinstance(name, str))
            except AssertionError:
                raise SchemaError("Name attribute should be a string type. Instead received: %s" % type(name))
            self.name = name

        if not subdir is None:
            self.subdir = subdir

    def __lt__(self, other):
        return len(self.name) < len(other.name)

    def __gt__(self, other):
        return len(self.name) > len(other.name)

    def __setattr__(self, attr, value):
        """
        This is an overloaded setter method to make sure that we're auto-expanding the filenames of files which exist.
        In the case we're assigning any other attributes the value is simply passed through
        Args:
            attr (str): This is the name of the attribute which we're assigning
            value (unknown): This is the value being assigned.
        """
        actual_value = value
        if attr == "name":
            actual_value = expandfilename(value)
        super(File, self).__setattr__(attr, actual_value)


    def _attribute_filter__set__(self, attribName, attribValue):
        if attribName is 'name':
            return expandfilename(attribValue)
        return attribValue

    def getPathInSandbox(self):
        """return a relative location of a file in a sandbox: subdir/name"""
        from GangaCore.Utility.files import real_basename
        return os.path.join(self.subdir, real_basename(self.name))

    def exists(self):
        """check if the file exists (as specified by 'name')"""
        import os.path
        return os.path.isfile(expandfilename(self.name))

    def create(self, outname):
        """create a file in  a local filesystem as 'outname', maintain
        the original permissions """
        import shutil

        shutil.copy(expandfilename(self.name), outname)
        if self.executable:
            chmod_executable(outname)

    def __repr__(self):
        """Get   the  representation   of  the   file.  Since   the  a
        SimpleStreamer uses  __repr__ for persistency  it is important
        to return  a valid python expression  which fully reconstructs
        the object.  """

        return "File(name='%s',subdir='%s')" % (self.name, self.subdir)

    def isExecutable(self):
        """  return true  if  a file  is  create()'ed with  executable
        permissions,  i.e. the  permissions of  the  existing 'source'
        file are checked"""
        return self.executable or is_executable(expandfilename(self.name))

# add File objects to the configuration scope (i.e. it will be possible to
# write instatiate File() objects via config file)
GangaCore.Utility.Config.config_scope['File'] = File

def string_file_shortcut_file(v, item):
    if isinstance(v, str):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        return File(v)
    return None

allComponentFilters['files'] = string_file_shortcut_file

class ShareDir(GangaObject):

    """Represents the directory used to store resources that are shared amongst multiple Ganga objects.

    Currently this is only used in the context of the prepare() method for certain applications, such as
    the Executable() application. A single ("prepared") application can be associated to multiple jobs.

    """
    _schema = Schema(Version(1, 0), {'name': SimpleItem(defvalue='', getter="_getName", doc='path to the file source'),
                                     'subdir': SimpleItem(defvalue=os.curdir, doc='destination subdirectory (a relative path)'),
                                     'associated_files': GangaFileItem(defvalue=[], typelist = [str, IGangaFile], doc='A list of files associated with the sharedir')})

    _category = 'shareddirs'
    _exportmethods = ['add', 'ls', 'path', 'remove', 'addAssociatedFile', 'listAssociatedFiles']
    _name = "ShareDir"

    def __init__(self, name=None, subdir=os.curdir):
        super(ShareDir, self).__init__()

        self._setRegistry(None)

        self._real_name = None

    def setSchemaAttribute(self, name, value):
        """
        This method intercepts the setter method used in reading in an XML file.
        This should be regarded as s specialist use case of overloading this method as
        this class plays fast and lose with the use of a getter for an attribute which is stored in memory.
        It may be worth reconsidering the use of a better getter method here if possible but unlikely as
        this would likely involve coding the class with implicit knowledge of it's proxy which is worse.
        If the Schema properly supported setters to match the getters this would be a much simplified structure.

        Args:
            name (str): Atribute being set
            value (unknown): value being assigned to that attribute
        """
        if name == "name":
            self._real_name = value
        else:
            # 'name' hidden behind setter here so shouldn't attempt to modify it
            super(ShareDir, self).setSchemaAttribute(name, value)

    def __setattr__(self, name, value):
        """
        A setattr wrapper which intercepts calls to assign self.name to self._real_name so the name gettter works
        Args:
            name (str): Name of the attribute being set
            value (unknown): value being assigned to the attribute
        """
        if name == 'name':
            self._real_name = value
        else:
            # 'name' hidden behind setter here so shouldn't attempt to modify it
            super(ShareDir, self).__setattr__(name, value)

    def __getattr__(self, name):
        """
        A getter wrapper which intercepts bare calls to name and redirects them to _real_name
        Args:
            name (str): The attribute which is being looked for
        """
        if name == 'name':
            return _getName()
        return super(ShareDir, self).__getattr__(name)

    def _getName(self):
        """
        A getter method for the 'name' schema attribute which will trigger the creation of a SharedDir on disk only when information about it is asked
        """

        with _prepare_lock:
            if not self._real_name:
                self._real_name = 'conf-{0}'.format(uuid.uuid4())
                share_dir = os.path.join(getSharedPath(), self._real_name)
                if not os.path.isdir(share_dir):
                    logger.debug("Actually creating: %s" % share_dir)
                    os.makedirs(share_dir)
                if not os.path.isdir(share_dir):
                    logger.error("ERROR creating path: %s" % share_dir)
                    raise GangaException("ShareDir ERROR")

        return self._real_name

    def add(self, input):
        from GangaCore.Core.GangaRepository import getRegistry
        if not isType(input, list):
            input = [input]
        for item in input:
            if isType(item, str):
                if os.path.isfile(expandfilename(item)):
                    logger.info('Copying file %s to shared directory %s' % (item, self.name))
                    shutil.copy2(expandfilename(item), os.path.join(getSharedPath(), self.name))
                    shareref = getRegistry("prep").getShareRef()
                    shareref.increase(self)
                    shareref.decrease(self)
                else:
                    logger.error('File %s not found' % expandfilename(item))
            elif isType(item, File) and item.name is not '' and os.path.isfile(expandfilename(item.name)):
                logger.info('Copying file object %s to shared directory %s' % (item.name, self.name))
                shutil.copy2(expandfilename(item.name), os.path.join(getSharedPath(), self.name))
                shareref = getRegistry("prep").getShareRef()
                shareref.increase(self)
                shareref.decrease(self)
            else:
                logger.error('File %s not found' % expandfilename(item.name))

    def path(self):
        """Get the full path of the ShareDir location"""
        return os.path.join(getSharedPath(), self._getName())
                
    def ls(self):
        """
        Print the contents of the ShareDir
        """
        full_shareddir_path = self.path()
        try:
            os.path.isdir(full_shareddir_path)
            cmd = "find '%s'" % (full_shareddir_path)
            files = os.popen(cmd).read().strip().split('\n')
            padding = '|  '
            for file in files:
                level = file.count(os.sep)
                level = level - 6
                pieces = file.split(os.sep)
                symbol = {0: '', 1: '/'}[os.path.isdir(file)]
                logger.info(padding * level + pieces[-1] + symbol)
        except IOError:
            logger.warn('ShareDir %s not found on storage' %
                        full_shareddir_path)

    def exists(self):
        """check if the file exists (as specified by 'name')"""
        import os.path
        return os.path.isdir(expandfilename(self.name))

    def create(self, outname):
        """create a file in  a local filesystem as 'outname', maintain
        the original permissions """
        import shutil

        shutil.copy(expandfilename(self.name), outname)
        if self.executable:
            chmod_executable(outname)

    def __repr__(self):
        """Get   the  representation   of  the   file.  Since   the  a
        SimpleStreamer uses  __repr__ for persistency  it is important
        to return  a valid python expression  which fully reconstructs
        the object.  """

        return "ShareDir(name='%s',subdir='%s')" % (self.name, self.subdir)

    def isExecutable(self):
        """  return true  if  a file  is  create()'ed with  executable
        permissions,  i.e. the  permissions of  the  existing 'source'
        file are checked"""
        return self.executable or is_executable(expandfilename(self.name))

    @synchronised
    def getAssociatedFiles(self):
        """ Load the list of associated files from the saved XML. This is
        necessary to keep the list consistent when copying jobs. """
        if os.path.isfile(os.path.join(self.path(), 'associated_files.xml')):
            with open(os.path.join(self.path(), 'associated_files.xml'), "r") as fobj:
                tmpobj, errs = from_file(fobj)
                self.associated_files = tmpobj

    @synchronised 
    def addAssociatedFile(self, newFile):
        """ Add an associated file to the ShareDir. Use this method to save
        it to theXML """
        self.getAssociatedFiles()
        self.associated_files.append(newFile)
        safe_save(os.path.join(self.path(), 'associated_files.xml'), self.associated_files, to_file)

    @synchronised
    def removeAssociatedFiles(self):
        """ Remove the files in the associated file list"""
        self.getAssociatedFiles()
        for entry in list(self.associated_files):
            if isinstance(entry, IGangaFile):
                entry.remove()
        self.associated_files = None

    @synchronised
    def listAssociatedFiles(self):
        """ List the associated_files of the ShareDir """
        self.getAssociatedFiles()
        return self.associated_files

    def remove(self):
        """ Remove the ShareDir and all of its associated files.
            This doesn't take care of the ShareRef so should only be
            called from there """
        shutil.rmtree(self.path(), ignore_errors=True)
        logger.info("Removed: %s" % self.path())
        self.removeAssociatedFiles()

GangaCore.Utility.Config.config_scope['ShareDir'] = ShareDir

def string_sharedfile_shortcut(v, item):
    if isinstance(v, str):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        return ShareDir(v)
    return None

allComponentFilters['shareddirs'] = string_sharedfile_shortcut


def cleanUpShareDirs():
    """Function to be used to clean up erronious empty folders in the Shared directory"""
    share_path = getSharedPath()

    logger.info("Cleaning Shared folders in: %s" % share_path)
    logger.info("This may take a few minutes if you're running this for the first time after 6.1.23, feel free to go grab a tea/coffee")

    for item in os.listdir(share_path):
        this_dir = os.path.join(share_path, item)
        if os.path.isdir(this_dir):
            # NB we do need to explicitly test the length of the returned value here
            # Checking is __MUCH__ faster than trying and failing to remove folders with contents on AFS
            if len(os.listdir(this_dir)) == 0:
                try:
                    os.rmdir(this_dir)
                except OSError:
                    logger.debug("Failed to remove: %s" % this_dir)

exportToGPI('cleanUpShareDirs', cleanUpShareDirs, 'Functions')

#
#
# $Log: not supported by cvs2svn $
# Revision 1.1  2008/07/17 16:40:53  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.15  2007/08/24 15:55:03  moscicki
# added executable flag to the file, ganga will set the executable mode of the app.exe file (in the sandbox only, the original file is not touched), this is to solve feature request #24452
#
# Revision 1.14  2007/01/25 16:18:21  moscicki
# mergefrom_Ganga-4-2-2-bugfix-branch_25Jan07 (GangaBase-4-14)
#
# Revision 1.12.2.4  2006/11/27 09:30:10  amuraru
# Extended the absolute paths detection in File objects
#
# Revision 1.12.2.3  2006/10/27 15:34:59  amuraru
# *** empty log message ***
#
# Revision 1.12.2.2  2006/10/27 15:33:59  amuraru
# bugs #20545
#
# Revision 1.12  2006/08/29 12:45:08  moscicki
# automatic conversion to absolute path
#
# Revision 1.11  2006/07/27 20:10:45  moscicki
# File to the configuartion symbol scope
#
# Revision 1.10  2006/02/17 14:21:32  moscicki
# fixed exists() method: using expanded file name
#
# Revision 1.9  2006/02/10 14:21:56  moscicki
# added exists() method
# docstring cleanup
#
# Revision 1.8  2005/08/23 17:07:23  moscicki
# Added executable flag for FileBuffer.
# Added create method for File and FileBuffer.
# Added getPathInSandbox() method.
#
# Revision 1.7  2005/08/10 09:45:36  andrew
# Added a subdir to File and FileBuffer objects. Changed the writefile method
# in FileWorspace to use the subdirectory
#
#
#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IMerger.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################
from GangaCore.GPIDev.Base.Proxy import GPIProxyObject
from GangaCore.Utility.Config import ConfigError, getConfig
from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException, IPostProcessor
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
import GangaCore.Utility.logging
import os

from GangaCore.GPIDev.Base.Proxy import isType
from posixpath import curdir, sep, pardir, join, abspath, commonprefix

logger = GangaCore.Utility.logging.getLogger()

def relpath(path, start=curdir):
    """Return a relative version of a path"""
    if not path:
        raise ValueError("no path specified")
    start_list = abspath(start).split(sep)
    path_list = abspath(path).split(sep)
    # Work out how much of the filepath is shared by start and path.
    i = len(commonprefix([start_list, path_list]))
    rel_list = [pardir] * (len(start_list) - i) + path_list[i:]
    if not rel_list:
        return curdir
    return join(*rel_list)

# set the mergers config up
config = getConfig("Mergers")

def getDefaultMergeDir():
    """Gets the default location of the mergers outputdir from the config"""

    try:
        config = getConfig('Mergers')
        outputdir = config['merge_output_dir']
    except ConfigError:
        outputdir = gangadir + "/merge_results"
    return os.path.expanduser(outputdir)


class IMerger(IPostProcessor):

    """
    Contains the interface for all mergers, all mergers should inherit from this object.
    """

    # set outputdir for auto merge policy flag
    # the default behaviour (True) is that outputdir is set by runAutoMerge() function in Merger.py module
    # however if this flag is set to False then merge() will be called for auto merge with sum_outputdir set to None
    # thus it is up to the subclass to decide where the output goes in case of
    # auto merge
    set_outputdir_for_automerge = True

    _category = 'postprocessor'
    _exportmethods = ['merge']
    _name = 'IMerger'
    _hidden = 1
    _schema = Schema(Version(1, 0), {
        'files': SimpleItem(defvalue=[], typelist=[str], sequence=1, doc='A list of files to merge.'),
        'ignorefailed': SimpleItem(defvalue=False, doc='Jobs that are in the failed or killed states will be excluded from the merge when this flag is set to True.'),
        'overwrite': SimpleItem(defvalue=False, doc='The default behaviour for this Merger object. Will overwrite output files.'),
    })
    order = 1

    __slots__ = list()

    def execute(self, job, newstatus):
        """
        Execute
        """
        if (len(job.subjobs) != 0):
            try:
                return self.merge(job.subjobs, job.outputdir)
            except PostProcessException as e:
                logger.error("%s" % e)
                return self.failure
        else:
            return True

    def merge(self, jobs, outputdir=None, ignorefailed=None, overwrite=None):

        if ignorefailed is None:
            ignorefailed = self.ignorefailed

        if overwrite is None:
            overwrite = self.overwrite

        from GangaCore.GPIDev.Lib.Job import Job

        if not outputdir:
            outputdir = getDefaultMergeDir()
        else:
            if isType(outputdir, Job):
                # use info from job
                outputdir = outputdir.outputdir
            else:
                outputdir = os.path.expanduser(outputdir)

        files = {}

        if isType(jobs, Job):
            if outputdir is None:
                outputdir = jobs.outputdir
            return self.merge(jobs.subjobs, outputdir=outputdir, ignorefailed=ignorefailed, overwrite=overwrite)

        if not len(jobs):
            logger.warning('The jobslice given was empty. The merge will not continue.')
            return self.success

        for j in jobs:
            # first check that the job is ok
            if j.status != 'completed':
                # check if we can keep going
                if j.status == 'failed' or j.status == 'killed':
                    if ignorefailed:
                        logger.warning('Job %s has status %s and is being ignored.', j.fqid, j.status)
                        continue
                    else:
                        raise PostProcessException('Job %s has status %s and so the merge can not continue. '
                                                   'This can be overridden with the ignorefailed flag.' % (j.fqid, j.status))
                else:
                    raise PostProcessException("Job %s is in an unsupported status %s and so the merge can not continue. '\
                    'Supported statuses are 'completed', 'failed' or 'killed' (if the ignorefailed flag is set)." % (j.fqid, j.status))

            if len(j.subjobs):
                sub_result = self.merge(
                    j.subjobs, outputdir=j.outputdir, ignorefailed=ignorefailed, overwrite=overwrite)
                if (sub_result == self.failure) and not ignorefailed:
                    raise PostProcessException('The merge of Job %s failed and so the merge can not continue. '
                                               'This can be overridden with the ignorefailed flag.' % j.fqid)

            import glob
            for f in self.files:

                for matchedFile in glob.glob(os.path.join(j.outputdir, f)):
                    relMatchedFile = ''
                    try:
                        relMatchedFile = os.path.relpath(
                            matchedFile, j.outputdir)
                    except Exception as err:
                        logger.debug("Err: %s" % err)
                        GangaCore.Utility.logging.log_unknown_exception()
                        relMatchedFile = relpath(matchedFile, j.outputdir)
                    if relMatchedFile in files:
                        files[relMatchedFile].append(matchedFile)
                    else:
                        files[relMatchedFile] = [matchedFile]

                if not len(glob.glob(os.path.join(j.outputdir, f))):
                    if ignorefailed:
                        logger.warning('The file pattern %s in Job %s was not found. The file will be ignored.', f, j.fqid)
                        continue
                    else:
                        raise PostProcessException('The file pattern %s in Job %s was not found and so the merge can not continue. '
                                                   'This can be overridden with the ignorefailed flag.' % (f, j.fqid))
                # files[f].extend(matchedFiles)

        for k in files.keys():
            # make sure we are not going to over write anything
            outputfile = os.path.join(outputdir, k)
            if os.path.exists(outputfile) and not overwrite:
                raise PostProcessException('The merge process can not continue as it will result in over writing. '
                                           'Either move the file %s or set the overwrite flag to True.' % outputfile)

            # make the directory if it does not exist
            if not os.path.exists(outputdir):
                os.makedirs(outputdir)

            # recreate structure from output sandbox
            outputfile_dirname = os.path.dirname(outputfile)
            if outputfile_dirname != outputdir:
                if not os.path.exists(outputfile_dirname):
                    os.mkdir(outputfile_dirname)

            # check that we are merging some files
            if not files[k]:
                logger.warning('Attempting to merge with no files. Request will be ignored.')
                continue

            # check outputfile != inputfile
            for f in files[k]:
                if f == outputfile:
                    raise PostProcessException(
                        'Output file %s equals input file %s. The merge will fail.' % (outputfile, f))
            # merge the lists of files with a merge tool into outputfile
            msg = None
            try:
                self.mergefiles(files[k], outputfile)

                # create a log file of the merge
                # we only get to here if the merge_tool ran ok
                log_file = '%s.merge_summary' % outputfile
                with open(log_file, 'w') as log:
                    log.write('# -- List of files merged -- #\n')
                    for f in files[k]:
                        log.write('%s\n' % f)
                    log.write('# -- End of list -- #\n')

            except PostProcessException as e:
                msg = str(e)

                # store the error msg
                log_file = '%s.merge_summary' % outputfile
                with open(log_file, 'w') as log:
                    log.write('# -- Error in Merge -- #\n')
                    log.write('\t%s\n' % msg)
                raise e

        return self.success



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

""" Refer to Ganga/PACKAGE.py for details on the purpose of this module.
"""

external_packages = {}
from GangaCore.Utility.Setup import PackageSetup

setup = PackageSetup(external_packages)


def standardSetup(setup=setup):
    import GangaCore.Utility.Setup
#    GangaCore.Utility.Setup.setPlatform('slc3_gcc323')

    for p in setup.packages:
        setup.prependPath(p, 'PYTHONPATH')
        setup.prependPath(p, 'PATH')
        setup.prependPath(p, 'LD_LIBRARY_PATH')



#!/usr/bin/env python
# =============================================================================
# $Id: summary.py,v 1.11 2010-04-30 13:50:50 rlambert Exp $
# =============================================================================
"""
*******************************************************************************
*        Class to contain the python xmlsummary object                        *
*        this simplifies the interface to the schema parser                   *
*        it can also be taken as an example of how to interact                *
*        with the xml schema/VTree object                                     *
*                                                                             * 
*        By necessity, the summary object hard-codes the name of several      *
*        of the tags and attributes of the LHCb schema.                       * 
*        This is required for merging and intelligent handling of the objects * 
*                                                                             * 
*                                                                             * 
* Print: returns an xml string                                                * 
* write: sends output to a file                                               * 
* set_step: is used to set 'initialize'/'finalize' etc.                       * 
* fill_file:    used to add information about a file                          * 
* fill_counter: used to create a counter in the xml                           * 
* fill_lumi:    used to create a counter to be stored under lumiCounters      * 
* fill_memory:  used to add memory usage information                          * 
* file_dict:    returns dictionary of input/output files by name and GUID     * 
* count_dict:   returns dictionary of counters/lumiCounters by name           * 
*******************************************************************************
"""
# =============================================================================
__author__ = 'Rob Lambert Rob.Lambert@cern.ch'
# =============================================================================

import os as __os__
#print __file__
__schema__= None

try:
    import imp as __imp__
    __schema__=__imp__.load_source('schema',__os__.sep.join(__file__.split(__os__.sep)[:-1]+['schema.py']))
    #print 'loaded from file'
except (ImportError, IOError):
    try:
        from XMLSummaryBase import schema as __schema__
    except ImportError:
        import schema as __schema__


#import the schema-handling
#def cleverImport(path=None):
#    global __schema__
#    if __schema__ is not None:
#        return
#    if path is None:
#try:
#    from XMLSummaryBase import schema as __schema__
#except ImportError:
#    import schema as __schema__
    #finally fail here if module cannot be found!
    
#    import imp
#    __schema__=imp.load_module('schema',__os__.path.expanduser(__os__.path.expandvars(path)))
#    return

#from __schema_module__ import *

__default_schema__='$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd'
__file_tag__ = 'file'
__count_tag__= 'counter'

# =============================================================================
# Simple algorithm which manipulates with counters 
# =============================================================================
class Summary(__schema__.VTree):
    """ Simple helper functions for writing the xml summary in python
    This class inherits from the VTree object, and has some hard coded element names
    merging might also be done here"""
    #a dictionary of name:object
    #__count_dict__={}
    #a dictionary of name:object
    #__file_dict__={}
    def __init__ ( self, schemafile=__default_schema__, construct_default=True ) :
        """ Constructor. Variables in schemafile are expanded.
        construct_default should always be left true unless you wish to parse a file straight away.
        module path optionally imports the schema module from a different path if it isn't already here"""
        #print 'crap'
        if 'Schema' not in str(type(schemafile)):
            #print str(type(schemafile))
            #print "I think it's a string"
            self.__schema__=__schema__.Schema(schemafile)
        else:
            self.__schema__=schemafile
        if construct_default:
            self.__element__=self.__schema__.create_default(self.__schema__.root()).__element__
        else:
            self.__element__=None
        self.__count_dict__={}
        for mother in self.__schema__.Tag_mothers(__count_tag__):
            self.__count_dict__[mother]={}
        self.__file_dict__={}
        for mother in self.__schema__.Tag_mothers(__file_tag__):
            self.__file_dict__[mother]={}
        super(Summary,self).__init__(element=self.__element__,schema=self.__schema__,mother=None,check=False)
        
    
    def __file_exists__(self, GUID, filename, isOutput=False):
        '''internal method, return the file if it already exists'''
        #mothers=self.__schema__.Tag_mothers(__file_tag__)
        self.file_dict()
        mother='input'
        if isOutput: mother='output'
        
        #if it exists, return it
        if (GUID is not None) and (GUID in self.__file_dict__[mother]):
            return self.__file_dict__[mother][GUID]
        elif (filename is not None) and (filename in self.__file_dict__[mother]):
            return self.__file_dict__[mother][filename]
        
        return None
    
    def __file_merger__(self, destination, filename=None, GUID=None, status=None, addevents=0, isOutput=False):
        '''merge two file objects'''
        mothers=self.__schema__.Tag_mothers(__file_tag__)
        mother='input'
        if isOutput: mother='output'
        
        #update filename or GUID if only one is present
        if (GUID is not None) and (filename is not None):
            if ( destination.attrib('name')==""):
                destination.attrib('name',filename)
                self.__file_dict__[mother][filename]=destination
            if ( destination.attrib('GUID')==""):
                destination.attrib('GUID',GUID)
                self.__file_dict__[mother][GUID]=destination
        
        #because it's a reference I can do this:
        if status is not None:
            if destination.__element__.attrib['status']=='fail':
                pass
                #a fail is a fail
            elif status=='fail':
                #a fail is a fail
                destination.__element__.attrib['status']=status
            elif destination.__element__.attrib['status']=='mult':
                #mult is mult!
                pass
            elif destination.__element__.attrib['status']=='full':# and status!='full':
                #more than one lot of this file!
                destination.__element__.attrib['status']='mult'
            elif status=='none':
                #none is default, don't overwrite anything
                pass
                #destination.attrib('status',status)
            else:
                destination.__element__.attrib['status']=status
        #print 'at end of merge for file', filename, 'adding', addevents, 'to', destination.value()
        #destination.value(destination.value()+addevents)
        #print destination.value()
        destination.__element__.text=str(int(float(destination.__element__.text))+int(addevents))
        return True
                
    def __fill_file__(self, filename=None, GUID=None, status=None, addevents=0, isOutput=False):
        '''Internal common method to fill the file processed information
        filename and GUID are the file to find/create
        status is a new status to set
        the number of events is taken from addevents'''
        if filename is "":
            filename=None
        if GUID is "":
            GUID=None
        if status is "":
            status=None
            
        if filename is None and GUID is None:
            return False
        
        mothers=self.__schema__.Tag_mothers(__file_tag__)
        mother='input'
        if isOutput: mother='output'
        
        if filename is not None:
            if "LFN:" not in filename.upper():
                if "PFN:" not in filename.upper():
                    filename="PFN:"+filename
        open_file=self.__file_exists__(GUID,filename,isOutput)

        #hardcoded tag for file entry :S
        
        if open_file is None:
                #file needs to be created
                open_file=self.__schema__.create_default(__file_tag__)
                #print 'default created'
                if (status==None): open_file.attrib('status','none')
                else: open_file.attrib('status',status)
                open_file.value(addevents)
                files=self.children('input')[0]
                if isOutput: files=self.children('output')[0]
                
                if filename is not None:
                    open_file.attrib('name',filename)
                    self.__file_dict__[mother][filename]=open_file
                if GUID is not None:
                    open_file.attrib('GUID',GUID)
                    self.__file_dict__[mother][GUID]=open_file
                files.__append_element__(open_file)
                return True
        
        return self.__file_merger__(open_file, filename,
                            GUID,status,
                            addevents, isOutput)
    def __recache__(self):
        self.counter_dict(True)
        self.file_dict(True)
                  
    def __buildcounter__(self, name, flag, nEntries=None, flag2=None, min=None,  max=None):
        '''method to build a VTree counter'''
        #pick the daughter to fill
        type=__count_tag__
        attrib="name"
        
        #if isLumi:
        #    if nEntries is not None: type="lumiStatEntity"
        #    else: type="lumiCounter"
        if nEntries is not None:
            type="statEntity"
        
        #create the xml object to fill
        cnt=self.__schema__.create_default(type)
        cnt.attrib(attrib,name)
        #if it's a simple counter...
        if nEntries is None: 
            cnt.value(flag)
            return cnt
        #else it's a statEntity
        val=[]
        for f in cnt.attrib('format'):
            if f=='Flag':
                val.append(flag)
            elif f=='Flag2':
                val.append(flag2)
            elif f=='Entries':
                val.append(nEntries)
        cnt.value(val)
        cnt.attrib('max',max)
        cnt.attrib('min',min)
        
        return cnt
    
    def fill_VTree_file(self, file, isOutput=False):
        '''append or merge a vtree file into the tree'''

        mothers=self.__schema__.Tag_mothers(__file_tag__)
        mother='input'
        if isOutput: mother='output'
        
        if file.__element__.attrib['name']!="":
            if "LFN:" not in file.__element__.attrib['name'].upper():
                if "PFN:" not in file.__element__.attrib['name'].upper():
                    file.__element__.attrib['name']="PFN:"+file.__element__.attrib['name']
        
        filename=file.__element__.attrib['name']
        GUID=file.__element__.attrib['GUID']
        open_file=self.__file_exists__(GUID,filename, isOutput)
        
        
        #if it doesn't exist, just add it
        if open_file is None:
                files=self.children(mother)[0]
                a=file.clone()
                files.__append_element__(a)
                self.__file_dict__[mother][filename]=a
                self.__file_dict__[mother][GUID]=a
                return True
        
        #else merge:
        return self.__file_merger__(open_file, filename,
                            GUID,file.__element__.attrib['status'],
                            int(file.__element__.text), isOutput)
        
    
    def isFailure(self):
        """the opposite of isSuccess, negates what is stored by the success tag"""
        return (self.children('success')[0].value()==False)
    
    def set_step( self, step="initialize", success=False ) :
        """ Set the step entry, and the success entry """
        test=True
        istep=self.children('step')[0]
        test= (istep.value(step) and test)
        
        #example setting success
        isuccess=self.children('success')[0]
        test= (isuccess.value(success) and test)
        
        return test
    
    def parse(self, xmlfile):
        """ parse an existing xml file, and load it into THIS object """
        self.__element__=self.__schema__.parse(xmlfile).__element__
        self.__recache__()
        
        
    def fill_output(self, filename=None, GUID=None, status=None, addevents=0):
        '''Method to fill the file processed information
        filename and GUID are the file to find/create
        status is a new status to set
        the number of events is taken from addevents'''
        return self.__fill_file__(filename,GUID,status,addevents,True)
    
    def fill_input(self, filename=None, GUID=None, status=None, addevents=0):
        '''Method to fill the file processed information
        filename and GUID are the file to find/create
        status is a new status to set
        the number of events is taken from addevents'''
        return self.__fill_file__(filename,GUID,status,addevents,False)
    
    def fill_VTree_counter(self, counter, isLumi=False):
        ''' fill a basic or complex stat counter, or lumi counter, from a VTree
        will merge the counter if it already exists.
        If you want to add, but not merge use the add method'''
        if 'VTree' not in str(type(counter)):
            raise TypeError('expected VTree, got '+str(type(counter))+' instead')
        if counter.tag() not in ['statEntity', __count_tag__]:
            raise TypeError('expected counter, got '+counter.tag()+' instead')
        mother='counters'#self.__schema__.Tag_mothers(counter.tag())[0]
        if isLumi:
            mother='lumiCounters'#self.__schema__.Tag_mothers(counter.tag())[1]

        #this takes a lot of time, why is it needed??
        #self.counter_dict(True)
        
        #hard coded, no way around it :S
        #attrib='name'
        name=counter.__element__.attrib['name']
        
        #check if it exists. If not, just add it
        #print self.__count_dict__
        #print self.__count_dict__[mother]
        if name not in self.__count_dict__[mother]:
            counters=__schema__.VTree([bt for bt in self.__element__.getchildren() if mother in bt.tag][0],self.__schema__,self,False)
            #counter=counter.clone()
            counters.__append_element__(counter)
            self.__count_dict__[mother][name]=counter
            return True
        
        #counters=self.children(mother)[0]
        #check if it exists. If not, just add it
        #cnt=counters.children(counter.tag(),attrib={attrib:name})
        #if cnt is None or len(cnt)==0:
        #    counters.add(counter)
        #    return True
        cnt=self.__count_dict__[mother][name]
        #else, merge the two
        if 'statentity' not in counter.tag().lower():
            cnt.__element__.text=str(int(cnt.__element__.text)+int(counter.__element__.text))
            return True
        #else be clever about the format
        val=cnt.value()
        for i in range(len(cnt.attrib('format'))):
            for j in range(len(counter.attrib('format'))):
                if cnt.attrib('format')[i]==counter.attrib('format')[j]:
                    #all entries are additive
                    val[i]=val[i]+counter.value()[j]
                    break
        cnt.value(val)
        if cnt.attrib('max')<counter.attrib('max'): cnt.attrib('max',counter.attrib('max'))
        if cnt.attrib('min')>counter.attrib('min'): cnt.attrib('min',counter.attrib('min'))
        
    def fill_counter(self, name, flag, nEntries=None, flag2=None, min=None, max=None):
        ''' fill a basic or complex stat counter '''
        cnt=self.__buildcounter__(name, flag, nEntries, flag2, min, max)
        return self.fill_VTree_counter(cnt, False)
    
    def fill_lumi(self, name, flag, nEntries=None, flag2=None, min=None, max=None):
        ''' fill a basic or complex Lumi stat counter '''
        cnt=self.__buildcounter__(name, flag, nEntries, flag2, min, max)
        return self.fill_VTree_counter(cnt, True)

    def fill_memory(self, memory, unit='b'):
        '''Method to fill the memory usage information'''
        usage=self.children('usage')[0]
        astat=usage.children('stat', attrib={'useOf':'MemoryMaximum'})
        if not astat or len(astat)==0:
            astat=self.__schema__.create_default('stat')
            astat.attrib('useOf','MemoryMaximum')
            astat.attrib('unit',unit)
            astat.value(memory)
            usage.add(astat)
        else:
            astat=astat[0]
            if astat.attrib('unit') != unit:
                raise AttributeError('I cannot compare two MemoryMaxima when they have different units! '+ astat.attrib('unit')+ " " +unit)
            if astat.value()<memory:
                astat.value(memory)
    
    def counter_dict(self, recache=False):
        '''return a name:VTree dictionary of counters'''
        mothers=self.__schema__.Tag_mothers(__count_tag__)
        empty=True
        if self.__count_dict__!={}:
            for mother in mothers:
                if self.__count_dict__[mother]!={}:
                    empty=False
        if empty or recache:
            for mother in mothers:
                self.__count_dict__[mother].clear()
                for amother in self.children(mother):
                    for count in amother.children():
                        self.__count_dict__[mother][count.attrib('name')]=count
        return self.__count_dict__
    
    def file_dict(self, recache=False):
        '''return a name:VTree dictionary of all input and output files'''
        mothers=self.__schema__.Tag_mothers(__file_tag__)
        empty=True
        if self.__file_dict__!={}:
            for mother in mothers:
                if self.__file_dict__[mother]!={}:
                    empty=False
        if empty or recache:
            for mother in mothers:
                self.__file_dict__[mother].clear()
                for amother in self.children(mother):
                    for file in amother.children():
                        if(file.attrib('name')!=""): self.__file_dict__[mother][file.attrib('name')]=file
                        if(file.attrib('GUID')!=""): self.__file_dict__[mother][file.attrib('GUID')]=file
        return self.__file_dict__

    
def Merge(summaries, schemafile=__default_schema__):
    '''Merge a list of summaries, return a new summary
    summaries can be a list of xml files to be parsed, or a list of summary objects'''
    if type("")==type(schemafile):
        schemafile=__schema__.Schema(schemafile)
    if type([]) != type(summaries): 
        raise TypeError('you should send a list into the merger, I got a ' + str(type(summaries)) + ' object instead')
    sum_objects=[]

    if type("")==type(summaries[0]) and str(type(Summary(schemafile=schemafile)))== str(type(summaries[0])):
        raise TypeError('you should send strings or Summaries into the merger, I got a ' + str(type(summaries[0])) + ' object instead')
    
    if str(type(Summary(schemafile=schemafile)))== str(type(summaries[0])):
        sum_objects=summaries
        #check they all have the same schema
        schema=sum_objects[0].__schema__.__schemafile_short__
        for asummary in sum_objects:
            if asummary.__schema__.__schemafile_short__ != schema:
                raise AttributeError('these xml files have different schema so cannot be merged')
    
    #try-except, in case you haven't given real summaries!
    
    try:
        
        if type("")==type(summaries[0]):
            for asummary in summaries:
                #print asummary
                sum_object=Summary(schemafile,construct_default=False)
                #don't check every summary, unless the merging fails
                #sum_object.parse(asummary)
                asummary=__os__.path.expanduser(__os__.path.expandvars(asummary))
                if not __os__.path.exists(asummary):
                    raise IOError('file does not exist '+str(asummary))
                sum_object.__element__=schemafile.__fast_parse__(asummary).__element__
                sum_object.__recache__()
                #print sum_object
                sum_objects.append(sum_object)
        
        #make default object
        merged=Summary(sum_objects[0].__schema__)
        #print 'made default object'
        #merge success
        #print 'merge success'
        flag = True
        for asummary in sum_objects:
            if asummary.isFailure():
                flag =False
                break
        merged.children('success')[0].value(flag)
        #merge step
        #print 'merge step'
        #enum will be in order
        steps=merged.__schema__.Tag_enumeration(merged.children('step')[0].tag())
        flag=''
        for asummary in sum_objects:
            if flag==asummary.children('step')[0].__element__.text: continue
            for step in steps:
                #retain the lowest possible step
                if flag==step: break
                if step==asummary.children('step')[0].__element__.text: 
                    flag=step
                    break
        merged.children('step')[0].__element__.text=flag
        #merge input/output, simple counters, usage
        #print 'merge ip/op simple counters, usage'
        counters={}
        #lumiCounters={}
        for asummary in sum_objects:
            #print sum_objects.index(asummary)
            #print 'usage'
            for stat in asummary.children('usage')[0].children('stat'):
                merged.fill_memory(stat.value(),stat.__element__.attrib['unit'])
            #print 'input'
            for file in asummary.children('input')[0].children():
                #merged.fill_VTree_file(file,False)
                merged.fill_input(file.__element__.attrib['name'],
                                  file.__element__.attrib['GUID'],
                                  file.__element__.attrib['status'],
                                  int(file.__element__.text))
            #print 'output'
            for file in asummary.children('output')[0].children():
                #merged.fill_VTree_file(file,True)
                merged.fill_output(file.__element__.attrib['name'],
                                  file.__element__.attrib['GUID'],
                                  file.__element__.attrib['status'],
                                  int(file.__element__.text))
            #merge counters
            #print 'counters'
            for cnt in asummary.children('counters')[0].children(__count_tag__):
                #print 'counter'
                #print cnt
                name=cnt.__element__.attrib['name']
                if name not in counters:
                    counters[name]=int(cnt.__element__.text)
                else:
                    counters[name]=(counters[name]+int(cnt.__element__.text))
                #if name not in merged.__count_dict__['counters']:
                #    cnt=cnt.clone()
                #merged.fill_VTree_counter(cnt)
            #merge counters
            #print 'lumiCounters'
            for cnt in asummary.children('lumiCounters')[0].children(__count_tag__):
                name=cnt.__element__.attrib['name']
                if name not in merged.__count_dict__['lumiCounters']:
                    cnt=cnt.clone()
                merged.fill_VTree_counter(cnt,isLumi=True)
        #merge collected counters
        for c in counters:
            merged.fill_counter(c,counters[c])
        #print 'merge statCounters'
        for asummary in sum_objects:
            #print sum_objects.index(asummary)
            #merge statCounters
            for cnt in asummary.children('counters')[0].children('statEntity'):
                merged.fill_VTree_counter(cnt.clone())
            #merge counters
            for cnt in asummary.children('lumiCounters')[0].children('statEntity'):
                merged.fill_VTree_counter(cnt.clone(), isLumi=True)
        if merged.__schema__.__check__(merged.__element__):
            return merged
        else:
            raise AttributeError('merged file could not be verified')
    except:
        print('Something went horribly wrong, I will now check all your summaries')
        for asummary in sum_objects:
            asummary.__schema__.__check__(asummary.__element__)
        raise
    return None



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import sys
from GangaCore.Core.exceptions import PluginError
from GangaCore.Utility.Shell import Shell
from GangaCore.Utility.logging import getLogger
from GangaLHCb.Lib.RTHandlers.LHCbGaudiRunTimeHandler import LHCbGaudiRunTimeHandler
from GangaLHCb.Lib.RTHandlers.LHCbGaudiDiracRunTimeHandler import LHCbGaudiDiracRunTimeHandler
import GangaCore.Utility.Config

from GangaGaudi.Lib.Applications.Gaudi import Gaudi

from GangaCore.GPIDev.Schema import SimpleItem

available_lhcb_apps = None
available_lhcb_packs = None
logger = getLogger()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def backend_handlers():
    backends = {'LSF': LHCbGaudiRunTimeHandler,
                'Interactive': LHCbGaudiRunTimeHandler,
                'PBS': LHCbGaudiRunTimeHandler,
                'SGE': LHCbGaudiRunTimeHandler,
                'Local': LHCbGaudiRunTimeHandler,
                'Condor': LHCbGaudiRunTimeHandler,
                'Remote': LHCbGaudiRunTimeHandler,
                'Dirac': LHCbGaudiDiracRunTimeHandler
                }
    return backends

def available_apps():
    global available_lhcb_apps
    if available_lhcb_apps is None:
        available_lhcb_apps = ["Gauss", "Boole", "Brunel",
                               "DaVinci", "Moore", "MooreOnline",
                               "Vetra", "Panoptes", "Erasmus",
                               "Alignment", "Noether", "Urania"]

    return available_lhcb_apps


def available_packs(appname=None):
    global available_lhcb_packs
    if available_lhcb_packs is None:
        available_lhcb_packs = {
            'Gauss': 'Sim',
            'Boole': 'Digi',
            'Brunel': 'Rec',
            'DaVinci': 'Phys',
            'Moore': 'Hlt',
            'MooreOnline': 'Hlt',
            'Vetra': 'Tell1',
            'Panoptes': 'Rich',
            'Bender': 'Phys',
            'Erasmus': '',
            'Noether': '',
            'Urania': 'PID',
            'Alignment': 'Alignment/Escher'
        }
    if appname is None:
        return available_lhcb_packs
    else:
        return available_lhcb_packs[appname]


def addNewLHCbapp(appname, use=''):
    temp_apps = available_apps()
    temp_packs = available_packs()
    try:
        assert isinstance(appname, str)
    except AssertionError:
        raise PluginError("Application name is not a string: %s" % str(appname))
    if any(str(appname).lower() == val.lower() for val in temp_apps):
        logger.warning("Error: %s is already in the list of supported apps, not adding" % appname)
        return
    global available_lhcb_apps
    global available_lhcb_packs
    available_lhcb_apps.append(str(appname))
    available_lhcb_packs[str(appname)] = use
    return


def available_versions(self, appname):
    """Provide a list of the available Gaudi application versions"""
    from . import EnvironFunctions
    return EnvironFunctions.available_versions(self, appname)


def guess_version(self, appname):
    """Guess the default Gaudi application version"""
    from . import EnvironFunctions
    return EnvironFunctions.guess_version(self, appname)


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def lumi(xmlsummary):
    '''given an XMLSummary object, will return the integrated luminosity'''
    #  print(xmlsummary.counter_dict()['lumiCounters']['IntegrateBeamCrossing/Luminosity'].value()[0],'+/-',xmlsummary.counter_dict()['lumiCounters']['IntegrateBeamCrossing/Luminosity'].value()[2])

    lumiDict = dict(list(zip(xmlsummary.counter_dict()['lumiCounters']['IntegrateBeamCrossing/Luminosity'].attrib('format'),
                        xmlsummary.counter_dict()['lumiCounters'][
        'IntegrateBeamCrossing/Luminosity'].value()
    ))
    )
    return '"%s +- %s"' % (lumiDict['Flag'], lumiDict['Flag2'])


def events(xmlsummary):
    '''given an XMLSummary object, will return the number of events input/output'''
    ad = xmlsummary.file_dict()
    evts = {}
    for type in ad.keys():
        if type not in evts:
            evts[type] = 0
        for file in ad[type].keys():
            if type == 'input' and ad[type][file].attrib('status') == 'mult':
                logger.warning(
                    'Warning, processed file ' + ad[type][file].attrib('name') + 'multiple times')
            if ad[type][file].attrib('GUID') == file:
                continue
            else:
                evts[type] += ad[type][file].value()
    return evts


def xmldatafiles(xmlsummary):
    '''return a dictionary of the files the xmlsummary saw as input'''
    returndict = {}
    for file in xmlsummary.file_dict()['input'].values():
        if file.attrib('status') in returndict:
            returndict[file.attrib('status')].update([file.attrib('name')])
        else:
            returndict[file.attrib('status')] = set([file.attrib('name')])
    return returndict


def xmldatanumbers(xmlsummary):
    '''return a dictionary of the number of files the xmlsummary saw as input'''
    returndict = {}
    for file in xmlsummary.file_dict()['input'].values():
        if file.attrib('status') in returndict:
            returndict[file.attrib('status')] = returndict[
                file.attrib('status')] + 1
        else:
            returndict[file.attrib('status')] = 1
    return returndict


def xmlskippedfiles(xmlsummary):
    '''get all skipped files from xml'''
    filedict = xmldatafiles(xmlsummary)
    skippedfiles = set()
    for stat in ['none', 'fail']:
        if stat in filedict:
            skippedfiles.update(filedict[stat])
    return skippedfiles


def activeSummaryItems():
    activeItems = {'lumi': lumi,
                   'events': events,
                   'xmldatafiles': xmldatafiles,
                   'xmldatanumbers': xmldatanumbers,
                   'xmlskippedfiles': xmlskippedfiles
                   }
    return activeItems




def getTaskByID(this_id):
    """returns proxied task accoding to this_id"""
    return getObjectByID(this_id, 'tasks')

def getJobByID(this_id):
    """returns proxied job according to this_id"""
    return getObjectByID(this_id, 'jobs')

def getObjectByID(this_id, reg_name):
    """returns proxies object from a repo based upon it's ID"""
    from GangaCore.Core.GangaRepository import getRegistryProxy
    jobs_reg = getRegistryProxy(reg_name)
    this_job = jobs_reg( this_id )
    return this_job

def makeRegisteredJob():
    """Makes a new Job and registers it with the Registry"""
    from GangaCore.GPIDev.Lib.Job.Job import Job
    j = Job()
    j._auto__init__()
    return j




"""Classes and functions that drive a robot run.

The Driver class encapsulates a run.

The loaddriver() factory method creates a new driver based on the Robot
configuration.

"""

from GangaRobot.Framework import Utility
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaRobot.Framework.exceptions import * #import Fatal, Break, Continue excpetions
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.Config import getConfig
import time

logger = getLogger()


class Driver(object):
    
    """Driver responsable for executing the actions of a run.
    
    The class contains the run list, action map and repeat options as well as
    the dorun() method to execute a run.
    
    An instance is typically created by using the module factory method
    loaddriver() but can also be created directly.
    
    The driver instance is not thread-safe, i.e. it is not safe to modify its
    attributes or call its methods from more than one thread at a time. In
    particular, you should not modify any attributes while a run is in progress.
    Instead a new instance should be created.
    
    All actions are executed in the context of a Ganga session. A new action
    instance is created for each execution of the action.
    
    """
    
    def __init__(self, run, actions, repeat = False):
        """Create a new driver.
        
        Keyword arguments:
        run -- A list of action names and sleep periods (seconds).
        actions -- A dictionary of action names to IAction classes.
        repeat -- Optional boolean indicating if the run should repeat
            indefinitely, defaults to False.
        
        Example:
        from GangaRobot.Lib.Core.CoreSubmitter import CoreSubmitter
        from GangaRobot.Lib.Core.CoreExtractor import CoreExtractor
        from GangaRobot.Lib.Core.CoreReporter import CoreReporter
        run = ['submit', 60, 'extract', 'report']
        actions = {'submit':CoreSubmitter,
                   'extract':CoreExtractor,
                   'report':CoreReporter}
        repeat = False
        driver = Driver(run, actions, repeat)
        
        """
        self.runid = None
        self.run = run
        self.actions = actions
        self.repeat = repeat

    def dorun(self):
        """Executes a run of actions and sleep periods.
        
        Initialises runid to the current UTC ID.
        
        """
        self.runid = Utility.utcid()
        while 1:
            logger.info("Start run %s with id '%s'.", self.run, self.runid)
            for action in self.run:
                try:
                    self._doaction(action)
                except GangaRobotContinueError as e:
                    logger.warning("Continue Error in Action '%s' with message '%s'. Run continued", action, e)
                except GangaRobotBreakError as e:
                    logger.warning("Break Error in Action '%s' with message '%s'. Run ended", action, e)
                    break
                except GangaRobotFatalError as e:
                    logger.error("Fatal Error in Action '%s' with message '%s'. Run aborted", action, e)
                    raise
                except Exception as e:
                    config = getConfig('Robot')
                    if (config['ExceptionBehaviour'] == 'Continue'):
                        logger.error("Error in Action '%s' with message '%s'. Run continued", action, e)
                    elif (config['ExceptionBehaviour'] == 'Break'):
                        logger.error("Error in Action '%s' with message '%s'. Run continued", action, e)
                        break
                    else:
                        logger.error("Abort run id '%s'. Action '%s' failed with message %s.", self.runid, action, e)
                        raise
            logger.info("Finish run id '%s'.", self.runid)
            if not self.repeat: 
                break

    def _doaction(self, action):
        """Executes the named action or sleep period.
        
        Keyword arguments:
        action -- The name of action or a sleep period (seconds).
        
        self.runid must be initialised.
        
        """
        assert self.runid
        if action in self.actions:
            # create new instance of iaction class
            iaction = self.actions[action]()
            logger.info("Execute action '%s'.", action)
            iaction.execute(self.runid)
        else:
            try:
                seconds = int(action)
                logger.info("Sleep for %d seconds.", seconds)
                time.sleep(seconds)
            except ValueError:
                logger.error("Unknown action '%s'.")
                raise


def _loadclass(fqcn):
    """Load the class identified by the fully-qualified class name."""
    #extract modulepath and classname
    if fqcn.count('.'):
        modulepath = fqcn[:fqcn.rfind('.')]
        classname = fqcn[fqcn.rfind('.')+1:]
    else:
        modulepath = '__main__'
        classname = fqcn
    #import module
    module = __import__(modulepath, globals(), locals(), [classname])
    #get class from module
    class_ = getattr(module, classname)
    #check type
    if not isinstance(class_, type):
        raise ValueError('%s is not a fully-qualified class name.' % fqcn)
    return class_


def loaddriver():
    """Create new driver based on Robot configuration options.

    Example of relevant configuration options:
    [Robot]
    Driver_Run = ['submit', 30, 'extract', 'report']
    Driver_Repeat = False
    Driver_Action_submit = GangaRobot.Lib.Core.CoreSubmitter.CoreSubmitter
    Driver_Action_extract = GangaRobot.Lib.Core.CoreExtractor.CoreExtractor
    Driver_Action_report = GangaRobot.Lib.Core.CoreReporter.CoreReporter
    
    """
    KEY_RUN = 'Driver_Run'
    KEY_REPEAT = 'Driver_Repeat'
    KEY_ACTION_PREFIX = 'Driver_Action_'
    
    config = Utility.getconfig()
    run = config[KEY_RUN]
    repeat = config[KEY_REPEAT]
    actions = {}
    #load action classes
    for key in config:
        if key.startswith(KEY_ACTION_PREFIX):
            action = key[len(KEY_ACTION_PREFIX):]
            fqcn = config[key]
            try:
                actions[action] = _loadclass(fqcn)
            except Exception as e:
                raise ApplicationConfigurationError("Cannot load class '%s'. Exception: '%s'" % (fqcn, e))
    #check actions exist for run
    for action in run:
        if not action in actions:
            try:
                int(action)
            except ValueError as e:
                raise ApplicationConfigurationError("Unknown action '%s'. Exception: '%s'" % (action, e))
            
            

    return Driver(run = run, actions = actions, repeat = repeat)



#!/usr/bin/env python
import sys
import os
import errno
import socket
import traceback
HOST = 'localhost'  # Standard loopback interface address (localhost)
PORT = int(sys.argv[1])        # Port to listen on
rand_hash = raw_input() 
import time
#We have to define an output function as a placeholder here.
def output(data):
    pass

#A function to shutdown an existing processes
def closeSocket():
    sc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sc.connect((HOST, PORT))
    sc.sendall(b'close server')
    sc.close()
end_trans = '###END-TRANS###'
#This is a wrapper for the client sockets
class socketWrapper(object):

    def __init__(self, skt):
        self._socket = skt

    def read(self):
        cmd = ''
        while end_trans not in cmd:
            data = self._socket.recv(1024)
            if not data:
                cmd = '###BROKEN###'
                break
            cmd += data
        #Check the random string is in the cmd so we know it came from a trusted source.
        if not rand_hash in cmd:
            return 'close-connection'
        if cmd == '###BROKEN###':
            return ''
        return_string = cmd.replace(rand_hash, '')
        return return_string.replace(end_trans, '')

#Start the socket
s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
s.bind((HOST, PORT))
s.listen(1024)
#Set 30 minute timeout
s.settimeout(1800)
while True:
    try:
        conn, addr = s.accept()
        sock = socketWrapper(conn)
        cmd = sock.read()
        #Here we define the output method to just send the output of the diracCommand wrapper.
        def output(data):
            conn.sendall(repr(data))

        if cmd=='close-connection':
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
            continue

        if cmd=='close-server':
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
            break
        try:
            print(eval(cmd))
        except:
            try:
                exec(cmd)
            except:
                print("Exception raised executing command (cmd) '%s'\n" % cmd)
                print(traceback.format_exc())

        conn.sendall('###END-TRANS###')
    #Catch the timeout and exit
    except socket.timeout:
        break

s.shutdown(socket.SHUT_RDWR)
s.close()



#!/usr/bin/env python
from queue import Queue
import threading


class DuplicateDataItemError(Exception):

    """
    Class raised when adding the same item in the Data object.
    """

    def __init__(self, message):
        self.message = message


class Data(object):

    """
    Class to define user dataset collection.
    """

    _attributes = ('collection', 'queue')

    def __init__(self, collection=None):
        if collection is None:
            collection = []

        self.collection = collection
        self.queue = Queue(maxsize=-1)
        self.lock = threading.Lock()

        for item in collection:
            self.queue.put(item)

    def getCollection(self):
        return self.collection

    def isEmpty(self):
        '''
        checks if the bounded queue is empty.
        '''
        return self.queue.empty()

    def addItem(self, item):
        '''
        try to put a new item in the queue. As the queue is defined with infinity number
        of slots, it should never throw "Queue.Full" exception.
        '''

        self.lock.acquire()
        try:
            if item not in self.collection:
                self.collection.append(item)

                self.queue.put(item)
            else:
                raise DuplicateDataItemError(
                    'data item \'%s\' already in the task queue' % str(item))
        finally:
            self.lock.release()

    def getNextItem(self):
        '''
        try to get the next item in the queue after waiting in max. 1 sec.

        if nothing available, the exception "Queue.Empty" will be thrown. 
        '''

        theItem = None

        self.lock.acquire()
        try:
            theItem = self.queue.get(block=True, timeout=1)
        finally:
            self.lock.release()

        return theItem



from GangaCore.Core.exceptions import ApplicationConfigurationError

import copy
import tempfile
import string

from GangaCore.Utility.Shell import Shell
from GangaCore.Utility.logging import getLogger

import GangaCore.Utility.Config
# Cannot configure LHCb without Gaudi changing underneath
gaudiConfig = GangaCore.Utility.Config.getConfig('GAUDI')

#---------------------------------------


def available_versions(self, appname):
    if self.newStyleApp is True:
        return available_versions_cmake(appname)
    else:
        return available_versions_SP(appname)


def guess_version(self, appname):
    if self.newStyleApp is True:
        return guess_version_cmake(appname)
    else:
        return guess_version_SP(appname)


def _getshell(self):
    if self.newStyleApp is True:
        return _getshell_cmake(self)
    else:
        return _getshell_SP(self)


def construct_merge_script(self, DaVinci_version, scriptName):
    if self.newStyleApp is True:
        return construct_merge_script_cmake(DaVinci_version, scriptName)
    else:
        return construct_merge_script_SP(DaVinci_version, scriptName)


def construct_run_environ(useCmake=False):
    if useCmake is True:
        return construct_run_environ_cmake()
    else:
        return construct_run_environ_SP()


def available_versions_cmake(appname):
    raise NotImplementedError


def guess_version_cmake(appname):
    raise NotImplementedError


def _getshell_cmake(self):
    raise NotImplementedError


def construct_merge_script(DaVinci_version, scriptName):
    raise NotImplementedError


def construct_run_environ_cmake():
    raise NotImplementedError


def construct_run_environ_SP():
    """
    This chunk of code has to run the SetupProject or equivalent at the start of
    a local/batch job's execution to setup the gaudi environment
    """
    script = """
# check that SetupProject.sh script exists, then execute it
os.environ['User_release_area'] = ''
#os.environ['CMTCONFIG'] = platform
f=os.popen('which SetupProject.sh')
setup_script=f.read()[:-1]
f.close()
if os.path.exists(setup_script):
    os.system('''/usr/bin/env bash -c '. `which LbLogin.sh` -c %s && source %s %s %s %s && printenv > env.tmp' ''' % (platform, setup_script,project_opts,app,version))
    for line in open('env.tmp').readlines():
        varval = line.strip().split('=')
        if len(varval) < 2:
            pass
        else:
            content = ''.join(varval[1:])
            # Lets drop bash functions
            if not str(content).startswith('() {'):
                os.environ[varval[0]] = content
    os.system('rm -f env.tmp')
else:
    sys.stdout.write('Could not find %s. Your job will probably fail.' % setup_script)
    sys.stdout.flush()
"""
    return script


def construct_merge_script_SP(DaVinci_version, scriptName):

    shell_script = """#!/bin/sh
SP=`which SetupProject.sh`
if [ -n $SP ]; then
  . SetupProject.sh  --force DaVinci %s
else
  echo "Could not find the SetupProject.sh script. Your job will probably fail"
fi
gaudirun.py %s
exit $?
""" % ( DaVinci_version, scriptName)

    script_file_name = tempfile.mktemp('.sh')
    try:
        script_file = open(script_file_name, 'w')
        script_file.write(shell_script)
        script_file.close()
    except:
        from GangaCore.Core.exceptions import PostProcessException
        raise PostProcessException('Problem writing merge script')

    return script_file_name


def available_versions_SP(appname):
    """Provide a list of the available Gaudi application versions"""
    s = Shell()
    tmp = tempfile.NamedTemporaryFile(suffix='.log')
    command = 'SetupProject.sh --ask %s' % appname
    rc, output, m = s.cmd1("echo 'q\n' | %s >& %s; echo" % (command, tmp.name))
    output = tmp.read()
    tmp.close()
    versions = output[output.rfind('(') + 1:output.rfind('q[uit]')].split()
    return versions.decode()


def guess_version_SP(appname):
    """Guess the default Gaudi application version"""
    s = Shell()
    tmp = tempfile.NamedTemporaryFile(suffix='.log')
    command = 'SetupProject.sh --ask %s' % appname
    rc, output, m = s.cmd1("echo 'q\n' | %s >& %s; echo" % (command, tmp.name))
    output = tmp.read()
    tmp.close()
    version = output[output.rfind(b'[') + 1:output.rfind(b']')]
    return version.decode()


def _getshell_SP(self):
    logger = getLogger()
    opts = ''
    if self.setupProjectOptions:
        opts = self.setupProjectOptions

    fd = tempfile.NamedTemporaryFile()
    script = '#!/bin/sh\n'
    if self.user_release_area:
        from GangaCore.Utility.files import expandfilename
        script += 'User_release_area=%s; export User_release_area\n' % \
            expandfilename(self.user_release_area)
    if self.platform:
        script += '. `which LbLogin.sh` -c %s\n' % self.platform
        #script += 'export CMTCONFIG=%s\n' % self.platform
    useflag = ''
    if self.masterpackage:
        from GangaLHCb.Lib.Applications.CMTscript import parse_master_package
        (mpack, malg, mver) = parse_master_package(self.masterpackage)
        useflag = '--use \"%s %s %s\"' % (malg, mver, mpack)
    cmd = '. SetupProject.sh %s %s %s %s' % (
        useflag, opts, self.appname, self.version)
    script += '%s \n' % cmd
    fd.write(script.encode())
    fd.flush()

    self.shell = Shell(setup=fd.name)
    if (not self.shell):
        raise ApplicationConfigurationError('Shell not created.')

    fd.close()

    app_ok = False
    identifier = "_".join([self.appname.upper(), self.version])
    for var in self.shell.env:
        if self.shell.env[var].find(identifier) >= 0:
            app_ok = True
            break
    if not app_ok:
        msg = 'Command "%s" failed to properly setup environment.' % cmd
        logger.error(msg)
        raise ApplicationConfigurationError(msg)

    self.env = copy.deepcopy(self.shell.env)

    return self.shell.env






from GangaCore.GPIDev.Base import GangaObject
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaCore.GPIDev.Lib.Job import MetadataDict
from GangaCore.GPIDev.Base.Proxy import stripProxy
from GangaCore.GPIDev.Lib.Tasks.common import getJobByID
import time

logger = getLogger()

########################################################################
def addInfoString( task_obj, info_str ):
    """Helper function to add an info string with a timestamp"""
    if len(task_obj.info) > 0 and task_obj.info[-1].find(info_str) > -1:
        # repeated string. Add one to instances
        num_rpts = 1
        if task_obj.info[-1].find("(rpt ") > -1:
            num_rpts = int( task_obj.info[-1][ task_obj.info[-1].find("(rpt ") + 5:-1 ] )
            
        task_obj.info[-1] = "%s: %s (rpt %i)" % (time.ctime(), info_str, num_rpts + 1)
        return
    
    task_obj.info.append("%s: %s" % (time.ctime(), info_str) )

class ITask(GangaObject):

    """This is the framework of a task without special properties"""
    _schema = Schema(Version(1, 0), {
        'transforms': ComponentItem('transforms', defvalue=[], sequence=1, copyable=0, doc='list of transforms'),
        'id': SimpleItem(defvalue=-1, protected=1, doc='ID of the Task', typelist=[int]),
        'name': SimpleItem(defvalue='NewTask', copyable=1, doc='Name of the Task', typelist=[str]),
        'comment': SimpleItem('', protected=0, doc='comment of the task', typelist=[str]),
        'status': SimpleItem(defvalue='new', protected=1, doc='Status - new, running, pause or completed', typelist=[str]),
        'float': SimpleItem(defvalue=0, copyable=1, doc='Number of Jobs run concurrently', typelist=[int]),
        'metadata': ComponentItem('metadata', defvalue=MetadataDict(), doc='the metadata', protected=1),
        'creation_date': SimpleItem(defvalue="19700101", copyable=0, protected=1, doc='Creation date of the task', typelist=[str]),
        'check_all_trfs': SimpleItem(defvalue=True, doc='Check all Transforms during each monitoring loop cycle'),
    })

    _category = 'tasks'
    _name = 'ITask'
    _exportmethods = ['run', 'appendTransform', 'overview', 'getJobs', 'remove', 'clone', 'pause', 'check', 'setBackend', 'setParameter',
                      'insertTransform', 'removeTransform', 'table', 'resetUnitsByStatus', 'removeUnusedJobs', 'n_all', 'n_status', 'n_all']

    _tasktype = "ITask"

    default_registry = "tasks"

# Special methods:
    def _auto__init__(self, registry=None):
        if registry is None:
            from GangaCore.Core.GangaRepository import getRegistry
            registry = getRegistry(self.default_registry)
        # register the job (it will also commit it)
        # job gets its id now
        registry._add(self)
        self.creation_date = time.strftime('%Y%m%d%H%M%S')
        self.startup()
        self.status = 'new'

    def startup(self):
        """Startup function on Ganga startup"""
        for t in self.transforms:
            t.startup()

    def getTransform(self, trf):
        """Get transform using either index or name"""
        if isinstance(trf, str):
            for trfid in range(0, len(self.transforms)):
                if trf == self.transforms[trfid].name:
                    return self.transforms[trfid]
            logger.warning("Couldn't find transform with name '%s'." % trf)
        elif isinstance(trf, int):
            if trf < 0 and trf > len(self.transforms):
                logger.warning("Transform number '%d' out of range" % trf)
            else:
                return self.transforms[trf]
        else:
            logger.warning('Incorrect type for transform referral. Allowed types are int or string.')

        return None

    def update(self):
        """Called by the monitoring thread. Base class just calls update on each Transform"""

        # if we're new, then do nothing
        if self.status == "new":
            return

        # loop over all transforms and call update
        for trf in self.transforms:
            if trf.status != "running":
                continue

            if trf.update() and not self.check_all_trfs:
                break

        # update status and check
        self.updateStatus()

# Public methods:
#
# - remove() a task
# - clone() a task
# - check() a task (if updated)
# - run() a task to start processing
# - pause() to interrupt processing
# - setBackend(be) for all transforms
# - setParameter(myParam=True) for all transforms
# - insertTransform(id, tf) insert a new processing step
# - removeTransform(id) remove a processing step

    def remove(self, remove_jobs="do_nothing"):
        """Delete the task"""

        # make sure the task isn't running
        if self.status.find("running") != -1:
            logger.error(
                "Task is still running. Please pause before removing!")
            return

        if not remove_jobs in [True, False]:
            logger.info("You want to remove the task %i named '%s'." %
                        (self.id, self.name))
            logger.info(
                "Since this operation cannot be easily undone, please call this command again:")
            logger.info(
                " * as tasks(%i).remove(remove_jobs=True) if you want to remove all associated jobs," % (self.id))
            logger.info(
                " * as tasks(%i).remove(remove_jobs=False) if you want to keep the jobs." % (self.id))
            return
        if remove_jobs:

            for trf in self.transforms:
                for unit in trf.units:
                    for jid in unit.active_job_ids:
                        try:
                            j = getJobByID(jid)
                            j.remove()
                        except Exception as err:
                            logger.debug("Remove Err: %s" % str(err))
                            pass

                    for jid in unit.prev_job_ids:
                        try:
                            j = getJobByID(jid)
                            j.remove()
                        except Exception as err2:
                            logger.debug("Remove Err2: %s" % str(err2))
                            pass

        self._getRegistry()._remove(self, auto_removed=1)
        logger.info("Task #%s deleted" % self.id)

    def clone(self):
        c = super(ITask, self).clone()
        for tf in c.transforms:
            tf.status = "new"
        c.check()
        return c

    def check(self):
        """This function is called by run() or manually by the user"""
        if self.status != "new":
            logger.error(
                "The check() function may modify a task and can therefore only be called on new tasks!")
            return
        try:
            for t in self.transforms:
                t.check()
        finally:
            self.updateStatus()
        return True

    def run(self):
        """Confirms that this task is fully configured and ready to be run."""
        if self.status == "new":
            self.check()

        if self.status != "completed":
            if self.float == 0:
                logger.warning(
                    "The 'float', the number of jobs this task may run, is still zero. Type 'tasks(%i).float = 5' to allow this task to submit 5 jobs at a time" % self.id)
            try:
                for tf in self.transforms:
                    if tf.status != "completed":
                        tf.run(check=False)

            finally:
                self.updateStatus()
        else:
            logger.info("Task is already completed!")

    def pause(self):
        """Pause the task - the background thread will not submit new jobs from this task"""
        float_cache = self.float
        self.float = 0
        if self.status != "completed":
            for tf in self.transforms:
                tf.pause()
            self.status = "pause"
        else:
            logger.info("Transform is already completed!")
        self.float = float_cache

    def insertTransform(self, id, tf):
        """Insert transfrm tf before index id (counting from 0)"""
        if self.status != "new" and id < len(self.transforms):
            logger.error(
                "You can only insert transforms at the end of the list. Only if a task is new it can be freely modified!")
            return
        # self.transforms.insert(id,tf.copy()) # this would be safer, but
        # breaks user exspectations
        # this means that t.insertTransform(0,t2.transforms[0]) will cause
        # Great Breakage
        self.transforms.insert(id, tf)
        stripProxy(tf).id = id

    def appendTransform(self, tf):
        """Append transform"""
        return self.insertTransform(len(self.transforms), tf)

    def removeTransform(self, id):
        """Remove the transform with the index id (counting from 0)"""
        if self.status != "new":
            logger.error("You can only remove transforms if the task is new!")
            return
        del self.transforms[id]

    def getJobs(self):
        """ Get the job slice of all jobs that process this task """
        jobslice = JobRegistrySlice("tasks(%i).getJobs()" % (self.id))
        for trf in self.transforms:
            for jid in trf.getJobs():
                jobslice.objects[getJobByID(jid).fqid] = stripProxy(getJobByID(jid))

        return JobRegistrySliceProxy(jobslice)

# Internal methods
    def updateStatus(self):
        """Updates status based on transform status.
           Called from check() or if status of a transform changes"""
        # Calculate status from transform status:
        states = [tf.status for tf in self.transforms]
        if "running" in states and "pause" in states:
            new_status = "running/pause"
        elif "running" in states:
            new_status = "running"
        elif "pause" in states:
            new_status = "pause"
        elif "new" in states:
            new_status = "new"
        elif "completed" in states:
            new_status = "completed"
        else:
            new_status = "new"  # no tranforms
        # Handle status changes here:
        if self.status != new_status:
            if new_status == "running/pause":
                logger.info("Some Transforms of Task %i '%s' have been paused. Check tasks.table() for details!" % (
                    self.id, self.name))
            elif new_status == "completed":
                logger.info("Task %i '%s' has completed!" %
                            (self.id, self.name))
            elif self.status == "completed":
                logger.warning(
                    "Task %i '%s' has been reopened!" % (self.id, self.name))
        self.status = new_status
        return self.status

    # Information methods
    def n_tosub(self):
        return self.float - sum([t.n_active() for t in self.transforms])

    def n_all(self):
        return sum([t.n_all() for t in self.transforms])

    def n_status(self, status):
        return sum([t.n_status(status) for t in self.transforms])

    def table(self):
        from GangaCore.Core.GangaRepository import getRegistryProxy
        t = getRegistryProxy('tasks').table(id=self.id)

    def overview(self, status=''):
        """ Show an overview of the Task """
        if status and not status in ['bad', 'hold', 'running', 'completed', 'new']:
            logger.error(
                "Not a valid status for unitOverview. Possible options are: 'bad', 'hold', 'running', 'completed', 'new'.")
            return

        print("Lists the units in each transform and give the state of the subjobs")
        print('')
        print(" " * 41 + "Active\tSub\tRun\tComp\tFail\tMinor\tMajor")
        for trfid in range(0, len(self.transforms)):
            print("----------------------------------------------------------------------------------------------------------------------")
            print("----   Transform %d:  %s" % (trfid, self.transforms[trfid].name))
            print('')
            self.transforms[trfid].overview(status)
            print('')

    def info(self):
        for t in self.transforms:
            t.info()

    def help(self):
        print("This is a Task without special properties")

    def resetUnitsByStatus(self, status='bad'):
        """Reset all units of the given status"""
        for trf in self.transforms:
            trf.resetUnitsByStatus(status)

    def removeUnusedJobs(self):
        """Remove any unused jobs"""
        for trf in self.transforms:
            trf.removeUnusedJobs()




__name__ = '__main__'
import os
import sys
import time
import datetime
import glob
import pickle
from functools import wraps
## NB parseCommandLine first then import Dirac!!
from DIRAC.Core.Base.Script import parseCommandLine
parseCommandLine()
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
dirac = DiracLHCb()

def diracCommand(f):
    '''
    This wrapper is intended to be used to wrap all 'commands' from the Ganga DIRAC API
    The intention is that all functions will now return a dict which is used to identify failures
    Args:
        f(function): Function we are wrapping
    '''
    @wraps(f)
    def diracWrapper(*args, **kwargs):
        ''' This method does the parsing of the wrapped function and it's output '''

        # When pipe_out == False this function is being called internally and shouldn't pipe the output to the streams
        if kwargs.get('pipe_out', True) is False:
            return f(*args, **kwargs)

        # We know we want to pipe the output to the streams when pipe_out == True
        output_dict = {}
        try:
            # Execute the function
            cmd_output = f(*args, **kwargs)
            if isinstance(cmd_output, dict) and 'OK' in cmd_output and ('Value' in cmd_output or 'Message' in cmd_output):
                # Handle the returned values from DIRAC HERE into a dictionary which Ganga can parse
                output_dict = cmd_output
            else:
                # Wrap all other returned objects in the output dict for Ganga
                output_dict['OK'] = True
                output_dict['Value'] = cmd_output
        except Exception as err:
            # Catch __ALL__ errors and report them back to Ganga
            # STDERR is lost in normal running so this will have to do!
            output_dict['OK'] = False
            output_dict['Message'] = str(err)

        # Pipe the output to the streams
        output(output_dict)

    return diracWrapper




'''Unique class to monitor and transfer dataset'''

import copy
import os
import subprocess
import sys
import types

from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Base import GangaObject 
from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import *
import GangaCore.Utility.logging

import db
import objectid
import utils


class SBDatasetManager(GangaObject):
    '''Methods to manage datasets.'''
    
    _name = "SBDatasetManager"
    _category = "datasetManager"
    _exportmethods = ["badDataset",
                      "closeDataset",
                      "createDataset",
                      "deleteDataset",
                      "downloadDataset",
                      "getFileList",
                      "openDataset",
                      "showDatasets",
                      "whoami"]
    _schema = Schema( Version( 1, 0 ), { })
    
    
    def createDataset(self):
        '''Interactive method to guide the user in dataset creation procedure.
        If the dataset is a 'personal production' type, force user to provide 
        a filter key.'''
        
        def asksParameter(parameter):
            '''Interactive method requesting user the value of each parameter 
            per session (FastSim, FullSim, Analysis)'''
            if parameter['customValue'] and len(parameter['values']) == 0:
                value = raw_input('\nEnter %s: ' % parameter['label'])
            elif not parameter['customValue'] and len(parameter['values']) == 0:
                raise GangaException('Invalid rule (customValue:False and values=0).')
            else:
                table = list()
                
                i = 0
                for value in parameter['values']:
                    table.append({'id': i, 'value': value})
                    i += 1
                
                if parameter['customValue']:
                    table.append({'id': i, 'value': 'Enter a custom value'})
                
                print('\nChoose %s:' % parameter['label'])
                column_names = ('id', 'value')
                print(utils.format_dict_table(table, column_names))
                index = utils.getIndex(maxExclusive=len(table))
                
                if parameter['customValue'] and index == len(table)-1:
                    value = raw_input('Custom value: ')
                else:
                    value = table[index]['value']
            
            # parameter insertion in dictionary. It will be subsequently 
            #inserted into dataset analysis bookkeeping table, hstore field
            new_dataset['parameters'][parameter['name']] = value
            
            return value
        
        
        type = [
            dict(id = 0, dataset_type = 'FastSim Personal Production'),
            dict(id = 1, dataset_type = 'FullSim Personal Production'),
            dict(id = 2, dataset_type = 'Analysis'),
            ]
        
        column_names = ('id', 'dataset_type')
        print(utils.format_dict_table(type, column_names))
        index = utils.getIndex(maxExclusive=len(type))
        
        new_dataset = dict()
        new_dataset['parameters'] = dict()
        
        ####################
        # FAST Simulation session
        ####################
        # parameter check: mandatory, free string param management
        # TODO: parameter type check, evaluate the config file option to store parameters
        
        if index == 0:
            new_dataset['session'] = 'fastsim'
            
            parameters = [
                {"name": "evt_file", "label": "Events per file", "customValue": True, "values": []},
                {"name": "analysis", "label": "Analysis", "customValue": True, "values": ["BtoKNuNu", "BtoKstarNuNu", "DstD0ToXLL", "DstD0ToXLL", "Generics", "HadRecoilCocktail", "KplusNuNu", "SLRecoilCocktail", "tau->3mu"]},
                {"name": "dg", "label": "Geometry", "customValue": True, "values": ["DG_4", "DG_4a", "DG_BaBar"]},
                {"name": "generator", "label": "Generator", "customValue": True, "values": ["B0B0bar_Btag-HD_Cocktail", "B0B0bar_Btag-SL_e_mu_tau_Bsig-HD_SL_Cocktail", "B0B0bar_generic", "B0B0bar_K0nunu", "B0B0bar_K0nunu_SL_e_mu_tau", "B0B0bar_Kstar0nunu_Kpi", "B0B0bar_Kstar0nunu_Kpi_SL_e_mu_tau", "B+B-_Btag-HD_Cocktail", "B+B-_Btag-SL_e_mu_tau_Bsig-HD_SL_Cocktail", "B+B-_generic", "B+B-_K+nunu", "B+B-_K+nunu_SL_e_mu_tau", "B+B-_Kstar+nunu", "B+B-_Kstar+nunu_SL_e_mu_tau", "B+B-_taunu_SL_e_mu_tau", "bhabha_bhwide", "ccbar", "tau+tau-_kk2f", "uds", "udsc", "Upsilon4S_generic"]},
                {"name": "bkg_mixing", "label": "Background Mixing Type", "customValue": True, "values": ["All", "NoPair", "NoMixing"]},
                {"name": "analysis_type", "label": "Analysis Type", "customValue": True, "values": ["BtoKNuNu", "BtoKstarNuNu", "HadRecoil", "SemiLepKplusNuNu"]}
            ]
            
            for parameter in parameters:
                asksParameter(parameter)
        
        ####################
        # FULL Simulation session
        ####################
        elif index == 1:
            new_dataset['session'] = 'fullsim'
            
            parameters = [
                {"name": "evt_file", "label": "Events per file", "customValue": True, "values": []},
                {"name": "sim_type", "label": "Simulation Type", "customValue": False, "values": ["fullsim", "background_frame"]},
                {"name": "generator", "label": "Generator", "customValue": False, "values": ["RadBhaBha", "singleparticle"]},
                {"name": "dg", "label": "Geometry", "customValue": True, "values": ["Geometry_CIPE", "Geometry_CIPE_BGO", "Geometry_CIPE_CSI", "Geometry_CIPE_V00-00-02"]},
                {"name": "pl", "label": "Physics list", "customValue": True, "values": ["QGSP", "QGSP_BERT", "QGSP_BERT_HP"]},
                {"name": "g4ver", "label": "Geant 4 version", "customValue": True, "values": ["9.2", "9.3"]},
                {"name": "opt_photons", "label": "Optical Photons", "customValue": False, "values": ["OFF", "ON"]}
            ]
            radbhabha = [
                {"name": "brunobbbminde", "label": "Min. Delta E", "customValue": True, "values": []}
            ]
            singleParticle = [
                {"name": "brunopdg", "label": "PDG Code", "customValue": True, "values": []},
                {"name": "brunothetamin", "label": "Theta min.", "customValue": True, "values": []},
                {"name": "brunothetamax", "label": "Theta max.", "customValue": True, "values": []},
                {"name": "brunophimin", "label": "Phi min.", "customValue": True, "values": []},
                {"name": "brunophimax", "label": "Phi max.", "customValue": True, "values": []},
                {"name": "brunoemin", "label": "Energy (GeV) min.", "customValue": True, "values": []},
                {"name": "brunoemax", "label": "Energy (GeV) max.", "customValue": True, "values": []}
            ]
            
            for parameter in parameters:
                value = asksParameter(parameter)
                
                # parameter dependencies management
                if parameter['name'] == 'generator':
                    if value == 'singleparticle':
                        parameters.extend(singleParticle)
                    elif value == 'RadBhaBha':
                        parameters.extend(radbhabha)
        
        ####################
        # ANALYSIS session
        ####################
        elif index == 2:
            new_dataset['session'] = 'analysis'
        else:
            raise GangaException('Invalid selection.')
        
        
        while True:
            free_string = raw_input('\nEnter free string: ')
            max_length = 128
            
            if len(free_string) <= max_length:
                new_dataset['parameters']['free_string'] = free_string
                break
            else:
                print('Free string must be <= %d char long.' % max_length)
        
        # dataset-site relation set
        new_dataset['site'] = getConfig('SuperB')['submission_site']
        new_dataset['owner'] = utils.getOwner()
        new_dataset['dataset_id'] = str(objectid.ObjectId())
        
        print('\nNew dataset details:')
        self.printDatasetDetail(new_dataset)
        
        
        value = ''
        while True:
            value = raw_input('Type \'yes\' to confirm the dataset creation or (q)uit: ')
            if value == 'yes':
                break
            elif value == 'q':
                raise utils.QuitException()
        
        sql = '''INSERT INTO analysis_dataset
            (owner, dataset_id, session, parameters, status)
            VALUES (%s, decode(%s, 'hex'), %s, %s, 'prepared');
            
            INSERT INTO analysis_dataset_site
            (dataset_id, site)
            VALUES (decode(%s, 'hex'), %s);'''
        params = (new_dataset['owner'], 
            new_dataset['dataset_id'], 
            new_dataset['session'], 
            new_dataset['parameters'],
            new_dataset['dataset_id'],
            new_dataset['site'])
        db.write(sql, params)
    
    def deleteDataset(self, **kwargs):
        '''to delete empty (prepared status) dataset'''
        kwargs['owner'] = utils.getOwner()
        kwargs['status'] = ['prepared']
        
        datasets = self.getDataset(**kwargs)
        dataset = self.printDatasets(datasets)
        dataset_id = dataset['dataset_id']
        
        sql = 'DELETE FROM analysis_dataset WHERE dataset_id = %s'
        db.write(sql, (r'\x' + dataset_id, ))
    
    def downloadDataset(self, **kwargs):
        '''to retrieve all files belonging to a owned dataset from GRID to 
        submission machine'''
        # TODO: create surl file lists beside the lfn list to permit lcg-cp 
        #fail over chain implamantation and to permit the direct plugin
        # subjob configuration by user given list
        
        kwargs['owner'] = utils.getOwner()
        kwargs['files'] = 0
        
        datasets = self.getDataset(**kwargs)
        dataset = self.printDatasets(datasets)
        
        dataset_id = dataset['dataset_id']
        files = dataset['files']
        occupancy_human = dataset['occupancy_human']
        
        home = os.path.expanduser('~')
        s = os.statvfs(home)
        free_disk = utils.sizeof_fmt_binary(s.f_bsize * s.f_bavail)
        
        #print('\nFree disk space: %s' % free_disk)
        print('\nTotal download size: %s\n' % occupancy_human)
        
        sql = 'SELECT lfn FROM analysis_output WHERE dataset_id = %s'
        lfns = db.read(sql, (r'\x' + dataset_id, ))
        
        localdir = os.path.join(home, dataset_id)
        os.mkdir(localdir)
        
        print('Downloading to %s ...' % localdir)
        i = 1
        
        for lfn in lfns:
            source = lfn['lfn']
            destination = os.path.join(localdir, source.split('/')[-1])
            
            process = subprocess.Popen(['lcg-cp', source, destination], stdout=subprocess.PIPE, close_fds=True)
            outData, errData = process.communicate()
            retCode = process.poll()
            
            if retCode != 0:
                raise Exception('lcg-cp fail with return code %d' % retCode)
            
            sys.stdout.write('\b' * 80 + '%s/%s' % (str(i), str(files)))
            sys.stdout.flush()
            
            i += 1
    
    def getFileList(self, **kwargs):
        '''It creates a <dataset_id>.txt with all LFN (or SRM) belonging to
         the chosen dataset'''
        kwargs['files'] = 0
        
        datasets = self.getDataset(**kwargs)
        dataset = self.printDatasets(datasets)
        dataset_id = dataset['dataset_id']
        
        sql = 'SELECT lfn FROM output_union WHERE dataset_id = %s'
        lfns = db.read(sql, (r'\x' + dataset_id, ))
        
        fileList = os.path.join(os.path.expanduser('~'), dataset_id + '.txt')
        
        # creating $home/<dataset_id>.txt
        try:
            f = open(fileList, 'w')
            for lfn in lfns:
                f.write("%s\n" % lfn['lfn'])
        finally:
            f.close()
        
        print('\nList %s created' % fileList)
    
    def badDataset(self, **kwargs):
        '''to set dataset status to bad'''
        kwargs['status'] = ['open', 'closed']
        self.__changeStatus('bad', **kwargs)
    
    def openDataset(self, **kwargs):
        '''to set dataset status to open'''
        kwargs['status'] = ['closed']
        self.__changeStatus('open', **kwargs)
    
    def closeDataset(self, **kwargs):
        '''to set dataset status to close
        TODO: kill unfinished job.'''
        kwargs['status'] = ['open']
        self.__changeStatus('closed', **kwargs)
    
    def __changeStatus(self, new_status, **kwargs):
        kwargs['owner'] = utils.getOwner()
        
        datasets = self.getDataset(**kwargs)
        dataset = self.printDatasets(datasets)
        dataset_id = dataset['dataset_id']
        
        sql = 'UPDATE analysis_dataset SET status = %s WHERE dataset_id = %s'
        db.write(sql, (new_status, r'\x' + dataset_id))
    
    def showDatasets(self, **kwargs):
        '''Print all metadata of dataset(s).
        Metodo esportato in GPI, usa massicciamente printDataset.'''
        
        def getParent(dataset):
            if 'parent' in dataset['parameters']:
                sql = 'SELECT * FROM dataset_union WHERE dataset_id = %s'
                parent = db.read(sql, (dataset['parameters']['parent'], ))
                
                assert len(parent) == 1, 'Must be only one parent.'
                getParent(parent[0])
                print('')
                self.printDatasetDetail(parent[0])
        
        datasets = self.getDataset(**kwargs)
        dataset = self.printDatasets(datasets)
        
        if dataset['session'] == 'analysis' and dataset['status'] not in ['prepared', 'temp']:
            print('\nParent datasets, older first:')
            getParent(dataset)
            
        print('\nSelected dataset:')
        self.printDatasetDetail(dataset)
    
    def printDatasetDetail(self, dataset):
        '''Print a table containing all the key-value dataset metadata'''
        
        # TODO: show the expiration time for dataset in temp status
        # to solve the nast dictionary problem (see hstore field model) all 
        # the dataset contents have been put to the same level
        
        d = copy.deepcopy(dataset)
        d.update(d['parameters'])
        del d['parameters']
        
        # dataset keys sorting and dictionary list creation
        items = d.items()
        items.sort()
        d = [{'key': key, 'value': value} for key, value in items]
        
        columns = ['key', 'value']
        print(utils.format_dict_table(d, columns))
    
    def printDatasets(self, datasets):
        ''' Given the heterogeneous dataset list, the method splits it in 
        categories and build the table per session. A unique id all over 
        the sessions permit the user to select univocally a dataset. All 
        the metadata will be printed with the parent chain: a parent dataset
        is defined as the one used as input to create the child one. The 
        method is public but not exported in GPI'''
        
        # check the term width
        (width, height) = utils.getTerminalSize()
        if width < 160:
            logger.error("Your terminal's width is %d; must be at least 160\nYou can make your font size smaller" % width)
            raise utils.QuitException()
        
        # better a named tuple but available in Python 2.6 only
        # fullsim: the table elements should be common to all the dataset keys in the list
        # print table rules:
        grouped_datasets = [
            # 0
            {'title': 'Fastsim Official Production',
             'dataset': list(),
             'order_by': ['prod_series', 'analysis', 'generator', 'dg', 'bkg_mixing', 'analysis_type'],
             'columns': ['id', 'prod_series', 'analysis', 'generator', 'dg', 'bkg_mixing', 'analysis_type', 'status']
             },
            # 1
            {'title': 'Fastsim Personal Production',
             'dataset': list(),
             'order_by': ['free_string', 'analysis', 'generator', 'dg', 'bkg_mixing', 'analysis_type'],
             'columns': ['id', 'free_string', 'analysis', 'generator', 'dg', 'bkg_mixing', 'analysis_type', 'status'],
             },
            # 2
            {'title': 'Fullsim Official Production',
             'dataset': list(),
             'order_by': ['prod_series', 'simtype', 'generator', 'dg', 'pl', 'g4ver', 'opt_photons'],
             'columns': ['id', 'prod_series', 'simtype', 'generator', 'dg', 'pl', 'g4ver', 'opt_photons', 'status']
             },
            # 3
            {'title': 'Fullsim Personal Production',
             'dataset': list(),
             'order_by': ['free_string', 'generator', 'dg', 'pl', 'g4ver', 'opt_photons'],
             'columns': ['id', 'free_string', 'generator', 'dg', 'pl', 'g4ver', 'opt_photons', 'status']
             },
            # 4
            {'title': 'Analysis',
             'dataset': list(),
             'order_by': ['free_string', 'creation_date'],
             'columns': ['id', 'free_string', 'creation_date', 'status']
             }
        ]
        
        for dataset in datasets:
            # put sub dictionary elements to level zero dictionary 
            for key, value in dataset.items():
                if type(dataset[key]) is types.DictType:
                    for key1, value1 in dataset[key].iteritems():
                        dataset[key1] = value1
                    #del dataset[key]
            
            # dataset selection
            if dataset['session'] == 'fastsim':
                if dataset['owner'] == 'Official':
                    grouped_datasets[0]['dataset'].append(dataset)
                else:
                    grouped_datasets[1]['dataset'].append(dataset)
            elif dataset['session'] == 'fullsim':
                if dataset['owner'] == 'Official':
                    grouped_datasets[2]['dataset'].append(dataset)
                else:
                    grouped_datasets[3]['dataset'].append(dataset)
            elif dataset['session'] == 'analysis':
                grouped_datasets[4]['dataset'].append(dataset)
            else:
                raise GangaException('session not recognized: %s' % dataset['session'])
        
        i = 0
        
        # field sort, adding id and print
        for group in grouped_datasets:
            if len(group['dataset']) > 0:
                print('\n%s' % group['title'])
                
                # dictionary sorting
                group['dataset'] = sorted(group['dataset'], key=lambda elem: ('%s ' * len(group['order_by'])) % tuple([elem[d] for d in group['order_by']]))
                
                # id adding
                for dataset in group['dataset']:
                    dataset['id'] = i
                    i += 1
                
                print(utils.format_dict_table(group['dataset'], group['columns']))
        
        
        # ask for input and print dataset details 
        if i == 1:
            index = 0
            print('\nAutomatically selected the only entry')
        else:
            print('\nChoose the dataset:')
            index = utils.getIndex(maxExclusive=i)
        
        
        # Object oriented solution to investigate and/or binary search
        # datasets have been grouped per print rules
        for group in grouped_datasets:
            for d in group['dataset']:
                if d['id'] == index:
                     for dataset in datasets:
                         if dataset['dataset_id'] == d['dataset_id']:
                            return dataset
    
    def getDataset(self, **kwargs):
        '''Get all metadata of all datasets.
        Public method, not exported to GPI.'''
        
        db_view_column = ['dataset_id', 'creation_date', 'occupancy']
        sql = 'SELECT * FROM dataset_union WHERE true'
        kwargs['owner'] = kwargs.get('owner', ['official', utils.getOwner()])
        
        # add filter to query
        if len(kwargs) > 0:
            for key, value in kwargs.iteritems():
                if key in db_view_column:
                    sql += " AND %s ILIKE '%s%%'" % (key, value)
                elif key == 'files':
                    sql += " AND files > %s" % value
                elif key in ['status', 'session', 'owner']:
                    if not isinstance(value, list):
                        value = [value]
                    
                    sql += " AND (false"
                    for s in value:
                        sql += " OR %s ILIKE '%s%%'" % (key, s)
                    sql += ")"
                    
                else:
                    sql += " AND parameters->'%s' ILIKE '%s%%'" % (key, value)
        
        # clean up the query
        sql = sql.replace('false OR ', '')
        sql = sql.replace('true AND ', '')
        
        # TODO: add control to prevent sql injection
        datasets = db.read(sql)
        
        if len(datasets) == 0:
            raise GangaException('No dataset found')
        
        i = 0
        for dataset in datasets:
            dataset['id'] = i
            i += 1
            dataset['occupancy_human'] = utils.sizeof_fmt_binary(dataset['occupancy'])
            if 'evt_file' in dataset['parameters'] and not 'evt_tot' in dataset['parameters']:
                evt_file = int(dataset['parameters']['evt_file'])
                if dataset['files'] is None:
                    dataset['files'] = 0
                files = int(dataset['files'])
                dataset['parameters']['evt_tot'] = evt_file * files
            if 'evt_tot' in dataset['parameters']:
                dataset['parameters']['evt_tot_human'] = utils.sizeof_fmt_decimal(int(dataset['parameters']['evt_tot']))
        
        return datasets
    
    def whoami(self):
        '''Print the User id string'''
        print(utils.getOwner())
    
logger = GangaCore.Utility.logging.getLogger()





from GangaCore.Utility.ColourText import getColour
import readline
import inspect
import re
import itertools
import subprocess
import types
import operator
from GangaCore.Utility.logging import getLogger
logger = getLogger('CompletionErrorChecker')


complete_bracket_matcher = re.compile('(\([^()]*?\))')
# used to not keep track of the replacements
# def bracket_sanitiser(text):


def bracket_sanitiser(text, replacement_list=None):
    if replacement_list is None:
        replacement_list = []
    #    print "checking:", text
    match_list = complete_bracket_matcher.findall(text)
    if match_list:
        for bracket in match_list:
            #            print "found:", bracket
            ##            text = text.replace(bracket,'')
            text = text.replace(bracket, '##%d##' % len(replacement_list), 1)
#            print "text:",text
            replacement_list.append(bracket)
#            print "rl:", replacement_list
##        text = bracket_sanitiser(text)
        text = bracket_sanitiser(text, replacement_list)
    return text

bracket_replacement_matcher = re.compile('##([0-9]+?)##')


def arg_splitter(text, strip_args=True):
    replacement_list = []
    text = bracket_sanitiser(text, replacement_list)
#    print "text:", text
#    print "rl:", replacement_list
    if not replacement_list or replacement_list[-1] == '()':
        return []
    if strip_args:
        # remove () and split and strip
        args = [x.strip() for x in replacement_list.pop()[1:-1].split(',')]
    else:
        args = replacement_list.pop()[1:-1].split(',')
#    print "args:", str(args)
    for i, a in enumerate(args):
        #        print "arg:", a
        while a.find('#') != -1:
            for r in bracket_replacement_matcher.findall(a):
                #                print "replacing:", r, str(replacement_list)
                a = a.replace('##%s##' % r, replacement_list[int(r)])
        args[i] = a
    return args

open_method_matcher = re.compile('([a-zA-Z0-9_.]+?)\(.*?')


def current_class_method_and_arg(text, namespace=globals()):
    '''
    should return class_object_label, class_name, method_name, arg_num
    '''
    text = bracket_sanitiser(text)
    matches = open_method_matcher.findall(text)
    namespace.update({'types': types})
#    print "open methods:", matches
    if not matches:
        return None, None, None, None
    split_match = matches[-1].rsplit('.', 1)  # get the class object as well

    class_label = None
    class_name = None
    method_name = split_match[-1]
    if len(split_match) > 1:
        class_label = split_match[0]
        class_name = eval('%s.__class__.__name__' % class_label, namespace)
        tmp = '.'.join([class_label, method_name])
        type_obj = eval('type(%s)' % tmp, namespace)
        if eval('%s != types.FunctionType and %s != types.MethodType' % (type_obj, type_obj), namespace):
            # constructor
            class_label = tmp
            class_name = eval('%s.__class__.__name__' % tmp, namespace)
            method_name = None
    elif eval('type(%s) != types.FunctionType and type(%s) != types.MethodType' % (method_name, method_name), namespace):
        # constructor
        class_label = method_name
        class_name = method_name
        method_name = None

    arg_count = text[text.find(split_match[-1]):].count(',')

    return class_label, class_name, method_name, arg_count


class GangaCompleter(object):

    def __init__(self, func, ns):
        self.func = func
        self.ns = ns
        self.ns['inspect'] = inspect

    def shell_size(self):
        '''
        Returns the size of the calling shell

        Returns a tuple of two ints representing the shells height and width respectively
        '''
        # note in python 3 can just do
        # shutil.get_terminal_size
        size_list = subprocess.Popen(
            ['stty', 'size'], stdout=subprocess.PIPE, stdin = subprocess.DEVNULL).communicate()[0].split()
        return eval(size_list[0]), eval(size_list[1])

    # want to push all colouring into the displayer as then wont mess up text
    # strngs for python < 2.whatever who dont get it
    def colouriser(self, match, user_input):
        #        try:
        if user_input.rstrip().endswith('(') or user_input.rstrip().endswith(','):
            class_label, class_name, method_name, arg_num = current_class_method_and_arg(
                user_input, self.ns)
            args = arg_splitter(match)
            var_args = [x for x in args if '*' in x]
            non_var_args = [x for x in args if '*' not in x]
            if method_name is None:  # constructor
                match = match.replace(
                    class_name, getColour('fg.green') + class_name + getColour('fg.normal'))
                return match

            match = match.replace(
                method_name, getColour('fg.green') + method_name + getColour('fg.normal'))
            if args:
                match = match.replace(', '.join(args), '##ARGS##')

            if arg_num < len(non_var_args):
                non_var_args[arg_num] = getColour(
                    'fg.blue') + non_var_args[arg_num] + getColour('fg.normal')
            else:
                if var_args:
                    var_args = [getColour('fg.blue') + x + getColour('fg.normal') for x in var_args]
                elif arg_num > 0:
                    logger.warning('Too many arguments provided')
                    match = match.replace(')', getColour(
                        'fg.red') + ')' + getColour('fg.normal'))

            return match.replace('##ARGS##', ', '.join(itertools.chain(non_var_args, var_args)))

        if user_input.endswith('.'):
            split_text = match.split('.')
            schema_items = {}
            if eval('hasattr(%s, "_schema") and hasattr(%s, "_readonly")' % (split_text[0], split_text[0]), self.ns):
                read_only = eval('%s._readonly()' % split_text[0], self.ns)
                schema_items = eval('dict(%s._schema.allItems())' % split_text[0], self.ns)

            elif eval('hasattr(%s,"_impl") and hasattr(%s._impl, "_schema") and hasattr(%s._impl, "_readonly")' % (split_text[0], split_text[0], split_text[0]), self.ns):
                read_only = eval('%s._impl._readonly()' % split_text[0], self.ns)
                schema_items = eval('dict(%s._impl._schema.allItems())' % split_text[0], self.ns)

            # if schema_items is not None and split_text[1] in schema_items:
            current_item = schema_items.get(split_text[1], None)
            if current_item is not None:
                if read_only:
                    split_text[1] = getColour(
                        'fg.red') + split_text[1] + getColour('fg.normal')
                elif current_item['protected']:
                    split_text[1] = getColour(
                        'fg.red') + split_text[1] + getColour('fg.normal')
                else:
                    split_text[1] = getColour(
                        'fg.green') + split_text[1] + getColour('fg.normal')
                return '.'.join(split_text)

        return match
#        except:
#            import traceback
#            print traceback.format_exc()

    def error_checker(self, user_input):
        # was going to use this to make displayer smaller by factoring out the
        # error checking bit
        pass

    # be on guard that the columns might break if the largest completion used to set the
    # column width is infact a method so gets added to by the colours
    def displayer(self, substitution, matches, longest_match_length):
        #        try:
        max_line_width = self.shell_size()[1]
        # max(itertools.imap(lambda x: len(x), matches)) + 2 #two blank spaces
        column_width = longest_match_length + 2
        num_per_line = max_line_width / column_width
        # need this as if partial auto-complete happens then get duplicate text
        # this is bcoz readline.get_line_buffer() already contains the partial
        # completed text
        user_input = readline.get_line_buffer()[:readline.get_endidx()]

        class_label, class_name, method_name, arg_num = current_class_method_and_arg(
            user_input, self.ns)

        if num_per_line == 0 or user_input.strip().endswith('(') or user_input.strip().endswith(','):
            num_per_line = 1

        # constructor
        if class_name is not None and method_name is None and user_input.strip().endswith(','):
            # if user_input.strip().endswith('(') or
            # user_input.strip().endswith(','):
            tmp = arg_splitter(user_input + ')', strip_args=False)
            new = tmp[:]  # map(lambda x: x.strip(), tmp)
            wrong = False
            unrecognised = []
            for i, arg in enumerate(filter(lambda x: x != '', map(lambda x: x.strip(), new))):
                if '=' in arg:
                    split_arg = arg.split('=')
                    if split_arg[0].strip() not in eval('dict(%s._impl._schema.allItems())' % class_name, self.ns):
                        unrecognised.append((split_arg[0].strip(), class_name))
                        # include = to avoid replacing letters e.g. a in the
                        # RHS also to keep the users spaces intact
                        new[i] = new[i].replace(split_arg[
                                                0] + '=', getColour('fg.red') + split_arg[0] + getColour('fg.normal') + '=')
                elif not eval('isinstance(%s, %s)' % (arg, class_name), self.ns):
                    new[i] = new[i].replace(
                        arg, getColour('fg.red') + arg + getColour('fg.normal'))
                    wrong = True

            user_input = user_input.replace(','.join(tmp), ','.join(new))
            if wrong:
                logger.warning('Only one positional arg allowed which must be an object of the same type to copy from')
                #user_input = user_input.replace(','.join(tmp), ','.join(new))
            if unrecognised:
                for a, c in unrecognised:
                    logger.warning(
                        "Unrecognised keyword argument, '%s' is not a modifyable attribute of '%s'" % (a, c))

        colour_green = getColour('fg.green')
        colour_red = getColour('fg.red')
        #colour_blue   = getColour('fg.blue')
        colour_normal = getColour('fg.normal')

        display_str = '\n'
        for i, m in enumerate(filter(lambda x: x != '', matches)):
            coloured_text = self.colouriser(m, user_input)
            width = column_width
            if colour_green in coloured_text:
                width = column_width + len(colour_green) + len(colour_normal)
            elif colour_red in coloured_text:
                width = column_width + len(colour_red) + len(colour_normal)
            display_str += ('{0:<%d}' % width).format(coloured_text)
            if num_per_line and (i + 1) % num_per_line == 0:
                display_str += '\n'

        if num_per_line and (i + 1) % num_per_line == 0:
            display_str += '\n'
        else:
            display_str += '\n\n'
        display_str += getColour('fg.blue')
        display_str += 'In [%s]:' % len(self.ns['In'])
        display_str += getColour('fg.normal')
        display_str += user_input

        print(display_str, end='')
        readline.redisplay()
#        except:
#            import traceback
#            print traceback.format_exc()

    def build_func_string(self, line):
        #        try:
        matches = ['']
        # get the argspec
        class_label, class_name, method_name, arg_num = current_class_method_and_arg(
            line, self.ns)
        if method_name is not None and class_name is not None:
            # must use class label here so as to get things not defined working
            # like queues
            tmp_arg_spec = eval(
                'inspect.getargspec(%s.%s)' % (class_label, method_name), self.ns)
            arg_spec = inspect.ArgSpec(tmp_arg_spec.args[
                                       1:], tmp_arg_spec.varargs, tmp_arg_spec.keywords, tmp_arg_spec.defaults)
        elif method_name is not None:
            arg_spec = eval('inspect.getargspec(%s)' % method_name, self.ns)
        elif class_name is not None:  # constructor
            schema_items = None
            if eval('hasattr(%s,"_schema")' % class_name, self.ns):
                schema_items = eval(
                    '%s._schema.allItems()' % class_name, self.ns)
            elif eval('hasattr(%s,"_impl") and hasattr(%s._impl,"_schema")' % (class_name, class_name), self.ns):
                schema_items = eval(
                    '%s._impl._schema.allItems()' % class_name, self.ns)

            # note wont work for build in types as these are defined in C or
            # things with slot wrappers
            if schema_items is None:
                tmp_arg_spec = eval(
                    'inspect.getargspec(%s.__init__)' % class_name, self.ns)
                arg_spec = inspect.ArgSpec(tmp_arg_spec.args[
                                           1:], tmp_arg_spec.varargs, tmp_arg_spec.keywords, tmp_arg_spec.defaults)
            else:
                t_args = []
                t_defaults = []
                for n, item in sorted(schema_items, key=operator.itemgetter(0)):
                    if not item._meta.get('hidden', False) and not item._meta.get('protected', False):
                        t_args.append(n)
                        t_defaults.append(item._meta.get('defvalue', None))
                if not t_defaults:
                    t_defaults = None
                else:
                    t_defaults = tuple(t_defaults)
                arg_spec = inspect.ArgSpec(t_args, None, None, t_defaults)
            matches.append('%s()' % class_name)
            matches.append('%s(%s)' % (class_name, class_name))
        # else:
        #    if class_name is not None:
        #        arg_spec     = eval('inspect.getargspec(%s.%s)' % (class_name, method_name), self.ns)
        #    else:
        #        arg_spec     = eval('inspect.getargspec(%s)' % method_name       , self.ns)

        # create string
        t = ''
        if method_name is not None and class_name is not None:
            t += '.'.join([str(class_name), str(method_name)])
        elif method_name is not None:
            t += str(method_name)
        elif class_name is not None:
            t += str(class_name)
        t += '('

        if arg_spec.defaults is None:
            t += ', '.join(arg_spec.args)
        else:
            num_default_free = len(arg_spec.args) - len(arg_spec.defaults)
            if arg_spec.args[:num_default_free]:
                t += ', '.join(arg_spec.args[:num_default_free]) + ', '
            for var, deflt in zip(arg_spec.args[num_default_free:], arg_spec.defaults):
                if isinstance(deflt, str):
                    t += "%s ='%s', " % (var, deflt)
                else:
                    t += '%s =%s, ' % (var, str(deflt))

        if arg_spec.varargs is not None:
            t += '*%s, ' % arg_spec.varargs
        if arg_spec.keywords is not None:
            t += '**%s, ' % arg_spec.keywords
        if t.endswith(', '):
            t = t[:-2]  # remove trailing ,
        t += ')'
        matches.append(t)
        return matches
#        except:
#            import traceback
#            print traceback.format_exc()

    def complete(self, text, state):
        #        try:
        response = None
        whole_line = readline.get_line_buffer()
        replace_complete_start = readline.get_begidx()
        replace_complete_end = readline.get_endidx()
        if state == 0:
            if text == ''\
                    and whole_line != '' \
                    and replace_complete_start == len(whole_line) \
                    and replace_complete_end == len(whole_line):
                #cls_label, class_name, method, arg_num = current_class_method_and_arg(whole_line.rstrip())
                # print cls, method
                if whole_line.rstrip().endswith('(') or whole_line.rstrip().endswith(','):
                    # if method is not None:
                    self.matches = self.build_func_string(whole_line.rstrip())

                # elif whole_line.rstrip().endswith(','):
                    # if method is not None:
                #    self.matches=self.build_func_string(whole_line.rstrip())
                else:  # e.g. a=<tab>
                    self.matches = [self.func(text, 0)]
                    for i in itertools.count(1):
                        next = self.func(text, i)
                        if next is None:
                            break
                        self.matches.append(next)
            else:
                self.matches = [self.func(text, 0)]
                for i in itertools.count(1):
                    next = self.func(text, i)
                    if next is None:
                        break
                    self.matches.append(next)
#        except:
#            import traceback
#            print traceback.format_exc()
#            raise
        try:
            response = self.matches[state]
        except IndexError:
            response = None
        return response



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 10/02/2014
################################################################################
"""@package ND280Executable
This module is designed to run any ND280 executable accessible in the $PATH environment variable.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re, time
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class ND280Executable(IPrepareApp):
    """
    ND280Executable application running any ND280 executables.

    The required input for this module are:
        app.exe = 'RunOARecon.exe'
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.outputfile = 'myFantasticResults.root'

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to analyze only 10 events:
        app.args = ['-n',10]
    """
    _schema = Schema(Version(1,1), {
        'exe' : SimpleItem(defvalue=None,typelist=['str', 'type(None)'],comparable=1,doc='A path (string) or a File object specifying an executable.'), 
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'outputfile' : SimpleItem(defvalue=None,doc='Output filename or filenames. Takes a string or a list of strings.', typelist=['str','type(None)','list']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'ND280Executable'
    _scriptname = None
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                  { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'outputfile', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                          { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'outputfile', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(ND280Executable,self).__init__()


    def configure(self,masterappconfig):
        
        self.args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup is None:
          raise ApplicationConfigurationError('No cmt setup script given.')

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in self.args:
          if arg == '-o':
            raise ApplicationConfigurationError('Option "-o" given in args. You must use the outputfile variable instead, even if you have multiple output files.')

        if self.outputfile is None:
          raise ApplicationConfigurationError('No output file given. Fill the outputfile variable.')
        else:
          if type(self.outputfile) == type([]):
            for OutFi in self.outputfile:
              self.args.append('-o')
              self.args.append(OutFi)
          else:
            self.args.append('-o')
            self.args.append(self.outputfile)


        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata is None:
          raise ApplicationConfigurationError('The inputdata variable is not defined.')
        fileList = job.inputdata.get_dataset_filenames()
        if len(fileList) < 1:
          raise ApplicationConfigurationError('No input data file given.')
        self.args.extend(fileList)

        argsStr = ' '.join(self.args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += self.exe+' '+argsStr+'\n'

        from GangaCore.GPIDev.Lib.File import FileBuffer

        if self.exe.find('.exe') > -1:
            scriptname = self.exe.replace('.exe', '.sh')
        else:
            scriptname = self.exe + '.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_ND280Executable') #_Properties
config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('ND280Executable','LSF', RTHandler)
allHandlers.add('ND280Executable','Local', RTHandler)
allHandlers.add('ND280Executable','PBS', RTHandler)
allHandlers.add('ND280Executable','SGE', RTHandler)
allHandlers.add('ND280Executable','Slurm', RTHandler)
allHandlers.add('ND280Executable','Condor', RTHandler)
allHandlers.add('ND280Executable','LCG', LCGRTHandler)
allHandlers.add('ND280Executable','gLite', gLiteRTHandler)
allHandlers.add('ND280Executable','TestSubmitter', RTHandler)
allHandlers.add('ND280Executable','Interactive', RTHandler)
allHandlers.add('ND280Executable','Batch', RTHandler)
allHandlers.add('ND280Executable','Cronus', RTHandler)
allHandlers.add('ND280Executable','Remote', LCGRTHandler)
allHandlers.add('ND280Executable','CREAM', LCGRTHandler)




#! /usr/bin/env python
import sys

doc = sys.argv[1]

f = file('Makefile-%s'%doc,'w')

tmpl = """
%(doc)sSRC := $(subst CVS,,$(wildcard src/%(doc)s/*))
%(doc)sTEX := $(filter %(TEXfilter)s,$(%(doc)sSRC))
%(doc)sRST := $(filter %(RSTfilter)s,$(%(doc)sSRC))
%(doc)sFIG := $(foreach fig,$(FIGS),$(wildcard src/%(doc)s/*.$(fig)))
.PHONY %(doc)s: %(doc)s.html %(doc)s.pdf %(doc)s.ps
.PHONY %(doc)s.html: html/%(doc)s/%(doc)s.html html/%(doc)s/%(doc)s.css
.PHONY %(doc)s.pdf: paper-$(PAPER)/%(doc)s.pdf
.PHONY %(doc)s.ps: paper-$(PAPER)/%(doc)s.ps
ALLCSSFILES += html/%(doc)s/%(doc)s.css
ALLHTML += %(doc)s.html
ALLPDF += %(doc)s.pdf
ALLPS += %(doc)s.ps
f_rst2html = $(foreach a,$(1),$(MKRSTHTML) $(a) $(subst src,html,$(subst .rst,.html,$(a)));)

html/%(doc)s/%(doc)s.html: $(%(doc)sSRC)
ifneq ($(%(doc)sTEX),)
	@echo $(%(doc)sTEX)
	mkdir -p html/%(doc)s
	$(MKHTML) --dir html/%(doc)s $(MKHTMLOPT) $(%(doc)sTEX)
endif
ifneq ($(%(doc)sRST),)
	mkdir -p html/%(doc)s
	$(call f_rst2html,$(%(doc)sRST))
endif
ifneq ($(strip $(%(doc)sFIG)),)
	cp  $(%(doc)sFIG) html/%(doc)s
	script/fiximgpath.py %(doc)s
endif
	rm -f html/%(doc)s/*.pl

paper-$(PAPER)/%(doc)s.pdf: $(%(doc)sSRC)
	mkdir -p paper-$(PAPER)
ifneq ($(strip $(%(doc)sFIG)),)
	cp $(%(doc)sFIG) paper-$(PAPER)
endif
	cd paper-$(PAPER) && $(MKPDF) ../src/%(doc)s/%(doc)s.tex

paper-$(PAPER)/%(doc)s.dvi: $(%(doc)sSRC)
	mkdir -p paper-$(PAPER)
ifneq ($(strip $(%(doc)sFIG)),)
	cp $(%(doc)sFIG) paper-$(PAPER)
endif
	cd paper-$(PAPER) && $(MKDVI) ../src/%(doc)s/%(doc)s.tex

paper-$(PAPER)/%(doc)s.ps: $(%(doc)sSRC)
	mkdir -p paper-$(PAPER)
ifneq ($(strip $(%(doc)sFIG)),)
	cp $(%(doc)sFIG) paper-$(PAPER)
endif
	cd paper-$(PAPER) && $(MKPS) ../src/%(doc)s/%(doc)s.tex
"""

print >> f, tmpl%{'doc':doc, 'TEXfilter':'%.tex', 'RSTfilter':'%.rst'}



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, GangaFileItem
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Lib.File.File import File
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList

class IVirtualization(GangaObject):

    """
    Contains the interface for all virtualization classes, all virtualization classes should inherit from this object.
    """

    def __init__(self, image):
        super(IVirtualization, self).__init__()
        self.image = image

    _schema = Schema(Version(1, 0), {
        'image': SimpleItem(defvalue="", typelist=[str], doc='Link to the container image'),
        'tokenuser': SimpleItem(defvalue="", typelist=[str], doc='Deploy token username'),
        'tokenpassword': SimpleItem(defvalue="", typelist=[str], doc='Deploy token password'),
        'mounts' : SimpleItem(defvalue={'/cvmfs': '/cvmfs'},
                               doc='Mounts to attempt from the host system. The key is the directory name on the host, and the value inside the container. If the directory is not available on the host, it will just be silently dropped from the list of mount points.'),
        'options': SimpleItem(defvalue=[], typelist=[list, GangaList], sequence=1, doc='A list of options to pass onto the virtualization command.')
    })
    _category = 'virtualization'
    _name = 'IVirtualization'
    _hidden = 1

    def modify_script(self, script):
        """Modify the given script by
        substituting virtualization related placeholders with relevant variables

            Arguments other than self:
               script - Script that need to be modified

            Return value: modified script"""

        return None



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ApplicationRuntimeHandlers.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

""" Application adapter table is a mechanism to match application and backend handlers.
"""


class _ApplicationRuntimeHandlers(object):

    def __init__(self):
        self.handlers = {}

    def add(self, application, backend, handler):
        self.handlers.setdefault(backend, {})[application] = handler

    def get(self, application, backend):
        return self.handlers[backend][application]

    def getAllBackends(self, application=None):
        if application is None:
            return list(self.handlers.keys())
        else:
            return [b for b in self.handlers.keys() if application in self.handlers[b]]

    def getAllApplications(self, backend=None):
        if backend is None:
            apps = {}

            for a in self.handlers.values():
                apps.update(a)

            return list(apps.keys())
        else:
            return list(self.handlers[backend].keys())

allHandlers = _ApplicationRuntimeHandlers()

if __name__ == '__main__':

    a = _ApplicationRuntimeHandlers()
    a.add('a', 'X', 1)
    a.add('a', 'Y', 1)
    a.add('b', 'X', 1)
    a.add('c', 'Z', 1)

    def compare(alist, blist):
        alist.sort()
        blist.sort()
        assert(alist == blist)

    compare(a.getAllBackends(), ['X', 'Y', 'Z'])
    compare(a.getAllApplications(), ['a', 'b', 'c'])

    compare(a.getAllBackends('a'), ['X', 'Y'])
    compare(a.getAllBackends('b'), ['X'])
    compare(a.getAllBackends('c'), ['Z'])

    compare(a.getAllApplications('X'), ['a', 'b'])
    compare(a.getAllApplications('Y'), ['a'])
    compare(a.getAllApplications('Z'), ['c'])



from GangaCore.Utility.Config import getConfig

import os
import cProfile
import time
import json

# creating a timestamp
timestr = time.strftime("-%Y%m%d-%H%M%S")

# Obtaining the config from .gangarc file
c = getConfig('Configuration')

if c['Profile_Memory']:
    from memory_profiler import profile
    
path = os.path.join(c['gangadir'], 'logs')


# A helper function to make the directories for storing log files
def _makedir(path):
    try:
        if not os.path.exists(path):
            os.makedirs(path)
        return path
    except Exception as e:
        raise e


# creating all the necessary directories
_makedir(path)
cpath = _makedir(os.path.join(path, 'cpu_profile/'))
mpath = _makedir(os.path.join(path, 'memory_profiles/'))
ccpath = _makedir(os.path.join(path, 'call_counters/'))


def cpu_profile(func):
    def wrapper(*args, **kwargs):
        # every function will have its different profile
        datafn = cpath + func.__name__ + timestr + ".profile"

        # profiling the function by passing in the function and arguments
        prof = cProfile.Profile()
        retval = prof.runcall(func, *args, **kwargs)

        # dump the stats to a file
        prof.dump_stats(datafn)
        return retval
    return wrapper


def cpu_profiler(cls, profile_cpu=c['Profile_CPU']):
    if not profile_cpu:
        pass
    else:
        for key, value in vars(cls).items():
            if callable(value):
                # adding cpu_profile decorator to every function of class
                setattr(cls, key, cpu_profile(value))
    return cls


def mem_profiler(cls, profile_memory=c['Profile_Memory']):
    if not profile_memory:
        pass
    else:
        file_name = mpath+cls.__name__+timestr+'.log'
        fp = open(file_name, 'w+')
        for key, value in vars(cls).items():
            if callable(value):
                # adding profile decorator to every function of class
                setattr(cls, key, profile(value, stream=fp, precision=6))
    return cls


function_calls = {}


def call_counts(func):
    def wrapper(*args, **kwargs):
        wrapper.calls += 1
        function_calls[(wrapper.__name__)] = wrapper.calls
        # storing the call counter for each function
        # Need to find a more effiecient way to store
        with open(ccpath+'call_counter_logs'+timestr+'.json', 'w+') as fp:
            json.dump(function_calls, fp, indent=4)
        return func(*args, **kwargs)
    wrapper.__name__ = func.__name__
    wrapper.calls = 0
    return wrapper


def call_counter(cls, count_calls=c['Count_Calls']):
    if not count_calls:
        pass
    else:
        for key, value in vars(cls).items():
            if callable(value):
                # adding call_counts decorator to every function of class
                setattr(cls, key, call_counts(value))
    return cls
